from collections import Counter, defaultdict, deque
import copy
import datetime
from typing import Iterable

from beancount.core.data import Transaction, Directive, iter_entry_dates, filter_txns
from beancount.ingest.extract import DUPLICATE_META


_Node = tuple[bool, Directive]  # (is_new_entry, entry)


def guess_transaction_duplicated(
        new_transaction: Transaction,
        existing_transaction: Transaction) -> bool:
    """Provides a rough guess that a new transaction already exists.
    
    Returns True iff for all the accounts that the new transaction describes,
    all postings under that account are the same in terms of units between
    the two transactions.
    """

    relevant_accounts = {
        posting.account
        for posting in new_transaction.postings
    }
    relevant_existing_data = Counter(
        (posting.account, posting.units)
        for posting in existing_transaction.postings
        if posting.account in relevant_accounts
    )
    relevant_new_data = Counter(
        (posting.account, posting.units)
        for posting in new_transaction.postings
    )

    return relevant_existing_data == relevant_new_data


def deduplicate(
        new_entries: list[Directive],
        existing_entries: list[Directive],
        window_days: int = 10) -> list[Directive]:
    """De-duplicate entries.

    A new non-transaction entry is considered connected to an existing entry
    iff they are identical.
    
    A new transaction is considered connected to an existing transaction iff:
    * the new transaction is no earlier than the existing transaction.
    * the new transaction is at most {window_days} days later than the existing
      transaction.
    * guess_transaction_duplicated returns True.

    If a new entry doesn't have any connection, it's considered non-duplicated.

    For each strongly connected subgraph, if all new entries are matched, all
    of them are considered duplicated. Otherwise, all of them are considered
    possibly duplicated.

    Returns new entries where:
    * Non-duplicated entries are preserved.
    * Duplicated entries are removed.
    * Possibly-duplicated entries are marked with DUPLICATE_META.
    """
    window_head = datetime.timedelta(days=window_days)
    window_tail = datetime.timedelta(days=1)

    matcher = _Matcher()
    for new_entry in new_entries:
        if isinstance(new_entry, Transaction):
            for existing_entry in filter_txns(iter_entry_dates(
                    existing_entries,
                    new_entry.date - window_head,
                    new_entry.date + window_tail)):
                if guess_transaction_duplicated(new_entry, existing_entry):
                    matcher.add_edge(id(new_entry), id(existing_entry))
        else:
            for existing_entry in iter_entry_dates(
                    existing_entries,
                    new_entry.date,
                    new_entry.date + window_tail):
                if new_entry == existing_entry:
                    matcher.add_edge(id(new_entry), id(existing_entry))
    
    duplicates: set[Directive] = set()
    possibly_duplicates: set[Directive] = set()

    matches = matcher.matches()
    for subgraph in matcher.subgraphs():
        n = len([True for node in subgraph if node in matches])
        if n == len(subgraph):  # duplicated
            duplicates.update(node[1] for node in subgraph if node[0])
        elif n:  # possibly duplicated
            possibly_duplicates.update(node[1] for node in subgraph if node[0])

    ret = []
    for new_entry in new_entries:
        if id(new_entry) in duplicates:
            continue
        elif id(new_entry) in possibly_duplicates and hasattr(new_entry, 'meta'):
            meta = copy.deepcopy(new_entry.meta)
            meta[DUPLICATE_META] = True
            ret.append(new_entry._replace(meta=meta))
        else:
            ret.append(new_entry)

    return ret


class _Matcher:
    def __init__(self) -> None:
        self._new_nodes: set[_Node] = set()
        self._edges: defaultdict[_Node, set[_Node]] = defaultdict(set)

    def add_edge(self, new_entry: Directive, existing_entry: Directive) -> None:
        new_node = (True, new_entry)
        existing_node = (False, existing_entry)
        self._new_nodes.add(new_node)
        self._edges[new_node].add(existing_node)
        self._edges[existing_node].add(new_node)

    def matches(self) -> set[_Node]:
        matches = set()
        for new_node in self._new_nodes:
            if new_node in matches:
                continue
            q: deque[_Node] = deque()
            q.append(new_node)
            visited = set()
            while q:
                u = q.popleft()
                if u in visited:
                    continue
                visited.add(u)
                if u not in matches and u != new_node:
                    matches.add(new_node)
                    matches.add(u)
                    break
                for v in self._edges[u]:
                    q.append(v)
        return matches

    def subgraphs(self) -> Iterable[set[_Node]]:
        visited = set()
        for new_node in self._new_nodes:
            if new_node in visited:
                continue
            subgraph_nodes = set()
            q: deque[_Node] = deque()
            q.append(new_node)
            while q:
                u = q.popleft()
                if u in visited:
                    continue
                subgraph_nodes.add(u)
                visited.add(u)
                for v in self._edges[u]:
                    q.append(v)
            yield subgraph_nodes



import os.path
from autobean.xcheck import plugin
import autobean.utils.plugin_test_utils as utils


@utils.generate_tests(os.path.dirname(__file__), plugin.CrossCheckPlugin.plugin)
def test() -> None:
    pass


def test_include_option() -> None:
    test_path = os.path.join(os.path.dirname(__file__), 'pass')
    source_path = os.path.join(test_path, 'source.bean')
    statement_path = os.path.join(test_path, '_statement.bean')
    ledger = utils.load_ledger(source_path)

    # assumptions
    assert not ledger.errors
    assert 'include' in ledger.options
    assert statement_path not in ledger.options['include']

    _, errors = utils.apply_plugin(plugin.CrossCheckPlugin.plugin, ledger.entries, ledger.options, None)

    assert not errors, errors
    assert statement_path in ledger.options['include'], 'statement files not added into options["include"]'



import collections
import decimal
import re
from typing import Any, Iterable, Iterator, Optional
from beancount.core import flags, account as account_lib
from beancount.core.data import Balance, Custom, Directive, Open, Pad, Transaction, new_metadata
from beancount.core.amount import Amount
from beancount.ops import pad, validation
from autobean.utils import error_lib, plugin_lib
from . import include, include_context, policy_lib, split_account

_DEFAULT_RECEIVABLE_ACCOUNT = 'Assets:Receivables'
_SUBACCOUNT_REGEX = re.compile(r':\[.*\]$')


@plugin_lib.plugin('autobean.share', param_type=str, custom_scope=r'autobean\.share($|\..*)')
class Plugin(plugin_lib.BasePlugin):

    def process(self, entries: list[Directive], options: dict[str, Any], arg: Optional[str]) -> Iterable[Directive]:
        self._enabled = True
        self._policy_db = policy_lib.PolicyDatabase()
        self._subaccounts = collections.defaultdict[str, set[str]](set)
        self._has_deferred_pad = False
        self._opened_accounts = set[str]()
        self._opened_account_ancestors = set[str]()
        self._receivable_account = _DEFAULT_RECEIVABLE_ACCOUNT
        errors = validation.validate(entries, options)
        self._error_logger.log_errors(errors)

        assert isinstance(arg, str)
        context = include_context.IncludeContext(viewpoint=arg)
        with include_context.try_enter_context(context) as effective_context:
            # process included files and links
            entries, errors = include.IncludePlugin.plugin(entries, options)
            self._error_logger.log_errors(errors)

            self._account_splitter = split_account.AccountSplitter(
                policy_db=self._policy_db,
                options=options,
                viewpoint=effective_context.viewpoint,
                asserted_accounts=get_asserted_accounts(entries))
            self._is_top_level = effective_context is context
            entries = list(super().process(entries, options, arg))
            if self._has_deferred_pad:
                entries, errors = pad.pad(entries, options)
                self._error_logger.log_errors(errors)
            if self._is_top_level:
                entries = self._account_splitter.process_open_close(entries)
            return entries

    def _check_enabled(self, entry: Custom) -> None:
        if not self._enabled:
            raise error_lib.PluginException(
                f'Cannot use {entry.type} while autobean.share is disabled')

    def _check_account_has_opened_descendants(self, account: str) -> None:
        if account not in self._opened_account_ancestors:
            raise error_lib.PluginException(
                f'Invalid reference to unknown account {account!r}')

    def _is_enabled(self) -> bool:
        return self._enabled

    def _is_receivable_account(self, account: str) -> bool:
        return (
            account == self._receivable_account or
            account.startswith(self._receivable_account + ':'))

    def _is_generated_account(self, account: str) -> bool:
        return (
            self._is_receivable_account(account) or
            re.search(_SUBACCOUNT_REGEX, account) is not None)

    def _open_account_if_not_opened(self, account: str, directive: Directive) -> Iterator[Open]:
        if account not in self._opened_accounts:
            self._opened_accounts.add(account)
            yield Open(
                meta=new_metadata(directive.meta['filename'], directive.meta['lineno']),
                date=directive.date,
                account=account,
                currencies=None,
                booking=None)

    # configs

    @plugin_lib.handle_custom('autobean.share.enable', 'exactly one bool')
    def handle_enable(self, entry: Custom, enable: bool) -> Iterable[Directive]:
        self._enabled = enable
        return ()

    @plugin_lib.handle_custom('autobean.share.receivable-account', 'exactly one account')
    def handle_receivable_account(self, entry: Custom, account: plugin_lib.Account) -> Iterable[Directive]:
        self._check_enabled(entry)
        self._receivable_account = account
        return ()

    @plugin_lib.handle_custom('autobean.share.policy', 'exactly one name or account')
    def handle_policy_def(self, entry: Custom, name: str | plugin_lib.Account) -> Iterable[Directive]:
        self._check_enabled(entry)
        policy_def = policy_lib.try_parse_policy_definition(entry.meta)
        if not policy_def:
            raise error_lib.PluginException('Empty share policy definition')
        self._policy_db.add_policy(name, policy_def)
        return ()

    # deferred directives

    @plugin_lib.handle_custom('autobean.share.balance', 'an account and an amount')
    def handle_deferred_balance(self, entry: Custom, account: plugin_lib.Account, amount: Amount) -> Iterator[Directive]:
        return self.handle_deferred_balance_tolerance(account, amount, None)

    @plugin_lib.handle_custom('autobean.share.balance', 'an account, an amount and a tolerance')
    def handle_deferred_balance_tolerance(self, entry: Custom, account: plugin_lib.Account | str, amount: Amount, tolerance: Optional[decimal.Decimal]) -> Iterator[Directive]:
        if not self._is_top_level:
            yield entry
            return
        self._check_enabled(entry)
        if not self._is_generated_account(account):
            raise error_lib.PluginException(f'autobean.share.balance must only be used on generated accounts')
        yield Balance(
            account=account,
            amount=amount,
            tolerance=tolerance,
        )

    @plugin_lib.handle_custom('autobean.share.pad', 'an account and an source account')
    def handle_deferred_pad(self, entry: Custom, account: plugin_lib.Account | str, source_account: plugin_lib.Account | str) -> Iterator[Directive]:
        if not self._is_top_level:
            yield entry
            return
        self._check_enabled(entry)
        if not self._is_generated_account(account) or not self._is_generated_account(source_account):
            raise error_lib.PluginException(f'autobean.share.pad must only be used on generated accounts')
        self._has_deferred_pad = True
        yield Pad(
            account=account,
            source_account=source_account,
        )

    # checks

    @plugin_lib.handle(Balance, when=_is_enabled)
    def handle_balance(self, entry: Balance) -> Iterator[Directive]:
        if self._is_generated_account(entry.account):
            raise error_lib.PluginException(
                f'balance must not be used on generated accounts. Consider using autobean.share.balance instead.')
        self._check_account_has_opened_descendants(entry.account)
        yield from self._account_splitter.process_balance(entry, self._error_logger)

    @plugin_lib.handle_custom('autobean.share.proportionate', 'exactly one account')
    def handle_proportionate(self, entry: Custom, account: plugin_lib.Account) -> Iterator[Directive]:
        self._check_enabled(entry)
        if self._is_generated_account(account):
            raise error_lib.PluginException(f'autobean.share.proportionate must not be used on generated accounts')
        self._check_account_has_opened_descendants(account)
        if custom := self._account_splitter.process_proportionate(entry, account):
            yield custom

    # transformations

    @plugin_lib.handle(Transaction, when=_is_enabled)
    def handle_transaction(self, entry: Transaction) -> Iterator[Directive]:
        if transaction := self._account_splitter.process_transaction(entry, self._receivable_account):
            for posting in transaction.postings:
                if self._is_receivable_account(posting.account):
                    yield from self._open_account_if_not_opened(posting.account, transaction)
            yield transaction

    @plugin_lib.handle(Pad, when=_is_enabled)
    def handle_pad(self, entry: Pad) -> Iterator[Directive]:
        # This should be rarely visited as pad is usually evaluated before this plugin.
        if self._is_generated_account(entry.account) or self._is_generated_account(entry.source_account):
            raise error_lib.PluginException(
                f'pad must not be used on generated accounts. Consider using autobean.share.pad instead.')
        yield entry

    @plugin_lib.handle(Open)
    def handle_open(self, entry: Open) -> Iterator[Directive]:
        yield entry
        self._opened_accounts.add(entry.account)
        for parent in account_lib.parents(entry.account):
            self._opened_account_ancestors.add(parent)
        if self._is_enabled():
            policy_def = policy_lib.try_parse_policy_definition(entry.meta)
            if policy_def:
                policy_lib.strip_share_meta(entry.meta)
                self._policy_db.add_policy(entry.account, policy_def)


def get_asserted_accounts(entries: Iterable[Directive]) -> set[str]:
    accounts = set()
    for entry in entries:
        if isinstance(entry, Balance):
            accounts.add(entry.account)
        elif (
                isinstance(entry, Custom) and
                entry.type == 'autobean.share.proportionate' and
                entry.values and
                isinstance(entry.values[0].value, str)):
            accounts.add(entry.values[0].value)
    return accounts



from typing import Any
from collections import defaultdict
from beancount.core.data import Directive, Custom
from autobean.utils import error_lib


class OutOfOrderDirectiveError(error_lib.Error):
    pass


def plugin(entries: list[Directive], options: dict[str, Any]) -> tuple[list[Directive], list[error_lib.Error]]:
    entries_by_file: defaultdict[str | None, list[Directive]] = defaultdict(list)
    ignored_files: set[str] = set()
    errors: list[error_lib.Error] = []
    for entry in entries:
        # Plugin-generated entries may not have associated file or line number.
        # We ignore entries in either case.
        if (hasattr(entry, 'date') and
                hasattr(entry, 'meta') and
                entry.meta.get('filename', None) and
                entry.meta.get('lineno', None)):
            should_append = True
            if is_enabling_directive(entry):
                if len(entry.values) != 1 or entry.values[0].dtype != bool:
                    errors.append(error_lib.InvalidDirectiveError(
                        entry.meta,
                        'autobean.sorted.enabled directive accepts a single '
                        'boolean argument',
                        entry))
                    should_append = False
            if should_append:
                entries_by_file[entry.meta['filename']].append(entry)

    # ignores entries with no filename or no line number
    entries_by_file.pop(None, None)
    for filename, file_entries in entries_by_file.items():
        if filename not in ignored_files:
            errors.extend(check_file_entries(file_entries))
    return entries, errors
    

def is_enabling_directive(entry: Directive) -> bool:
    return isinstance(entry, Custom) and entry.type == 'autobean.sorted.enabled'


def check_file_entries(entries: list[Directive]) -> list[error_lib.Error]:
    """Checks entries are in order and finds out-of-order entries.

    We find a longest non-descending subsequence and assumes all other
    entries are out-of-order.

    When finding the longest non-descending subsequence, we prefer the one
    whose maximum date is minimal. For example, in time sequence 1 2 100 3
    it's more likely that 1 2 3 is correct and 100 is a mistake (compared to
    1 2 100 as correct and 3 as mistake).
    """

    prevs = []
    # [(length, -maxdate)]
    scores: list[tuple[int, int]] = []

    sorted_entries = []
    enabled = True
    for entry in sorted(entries, key=lambda e: e.meta['lineno']):
        if is_enabling_directive(entry):
            enabled = entry.values[0].value
            continue
        if enabled:
            sorted_entries.append(entry)

    global_best_i = -1

    for entry in sorted_entries:
        best = (1, -entry.date.toordinal())
        prev = -1
        for i, score in enumerate(scores):
            if sorted_entries[i].date <= entry.date:
                current = (
                    score[0] + 1,
                    max(score[1], -sorted_entries[i].date.toordinal()))
                if current > best:
                    best = current
                    prev = i
        scores.append(best)
        prevs.append(prev)
        if global_best_i == -1 or best > scores[global_best_i]:
            global_best_i = len(scores) - 1
    prevs.append(global_best_i)

    misplaced_entries = []
    i = len(prevs) - 1
    while i >= 0:
        j = prevs[i]
        for k in range(i - 1, j, -1):
            misplaced_entries.append(sorted_entries[k])
        i = j

    errors: list[error_lib.Error] = []
    for misplaced_entry in misplaced_entries[::-1]:
        errors.append(OutOfOrderDirectiveError(
            misplaced_entry.meta,
            'Directive date does not follow non-descending order within the '
            'file',
            misplaced_entry))
    
    return errors



"""Shared context between including and included files."""

import collections
import contextlib
import dataclasses
import threading
from typing import DefaultDict, Iterator

_THREAD_LOCAL = threading.local()


@dataclasses.dataclass(frozen=True)
class IncludeContext:
    viewpoint: str
    subaccounts: DefaultDict[str, set[str]] = dataclasses.field(
        default_factory=lambda: collections.defaultdict(set))


@contextlib.contextmanager
def try_enter_context(context: IncludeContext) -> Iterator[IncludeContext]:
    original_context = getattr(_THREAD_LOCAL, 'include_context', None)
    try:
        if original_context:
            yield original_context
        else:
            _THREAD_LOCAL.include_context = context
            yield context
    finally:
        _THREAD_LOCAL.include_context = original_context



import collections
import dataclasses
import datetime
import decimal
import re
from typing import Any, Iterable, Optional
from beancount.core import amount
from beancount.core.data import Custom, Directive, Transaction, filter_txns, iter_entry_dates
from autobean.utils import error_lib


_MAIN_ACCOUNT_REGEX = re.compile(r':\[.*\]$')


@dataclasses.dataclass(frozen=True)
class Link:
    path: str
    account: str
    complement_path: str
    complement_account: str
    custom: Custom


class UnresolvedLinkError(error_lib.Error):
    pass


def link_accounts(
        entries_by_file: dict[str, list[Directive]],
        links: Iterable[Link],
        logger: error_lib.ErrorLogger) -> list[Directive]:

    _check_links(entries_by_file, links, logger)
    edges = _build_graph(entries_by_file, links, logger)
    return _resolve_links(entries_by_file, edges, logger)


def _main_account(account: str) -> str:
    return re.sub(_MAIN_ACCOUNT_REGEX, '', account)


def _check_links(
        entries_by_file: dict[str, list[Directive]],
        links: Iterable[Link],
        logger: error_lib.ErrorLogger) -> None:
    all_endpoints = set()
    for link in links:
        endpoints = [
            (link.path, link.account),
            (link.complement_path, link.complement_account),
        ]
        for ep in endpoints:
            if ep in all_endpoints:
                logger.log_error(error_lib.InvalidDirectiveError(
                    link.custom.meta,
                    f'Account {ep[1]} in {ep[0]} has multiple links',
                    link.custom,
                ))
            if ep[0] not in entries_by_file:
                logger.log_error(error_lib.InvalidDirectiveError(
                    link.custom.meta,
                    f'Ledger {ep[0]} was not included with '
                    f'"autobean.share.include" from this ledger',
                    link.custom,
                ))
            all_endpoints.add(ep)


def _build_graph(
        entries_by_file: dict[str, list[Directive]],
        links: Iterable[Link],
        logger: error_lib.ErrorLogger) -> dict[int, list[tuple[Transaction, str]]]:

    day = datetime.timedelta(days=1)
    edges = collections.defaultdict(list) # id(txn) -> [(complement txn, account)]
    for link in links:
        entries = entries_by_file.get(link.path, [])
        complement_entries = entries_by_file.get(link.complement_path, [])

        for entry in filter_txns(entries):
            expected_complement_feature = _transaction_feature(
                entry, link.account, True)
            if not expected_complement_feature[1]:
                continue
            # find complement txn and check duplicated complement
            complement_txn = None
            complement_duplicated = False
            for complement_entry in filter_txns(iter_entry_dates(
                    complement_entries, entry.date, entry.date + day)):
                complement_feature = _transaction_feature(
                    complement_entry, link.complement_account, False)
                if complement_feature == expected_complement_feature:
                    if not complement_txn:
                        complement_txn = complement_entry
                    else:
                        complement_duplicated = True
                        break
            # check duplicated
            duplicated = False
            found = False
            feature = _transaction_feature(entry, link.account, False)
            for entry2 in filter_txns(iter_entry_dates(
                    entries, entry.date, entry.date + day)):
                feature2 = _transaction_feature(entry2, link.account, False)
                if feature2 == feature:
                    if found:
                        duplicated = True
                        break
                    found = True
            if not complement_txn:
                logger.log_error(UnresolvedLinkError(
                    entry.meta,
                    f'No complement transaction found for link {link}',
                    entry,
                ))
            elif complement_duplicated:
                logger.log_error(UnresolvedLinkError(
                    entry.meta,
                    f'Multiple complement transactions found for link {link}',
                    entry,
                ))
            elif duplicated:
                logger.log_error(UnresolvedLinkError(
                    complement_txn.meta,
                    f'Multiple complement transactions found for link {link}',
                    complement_txn,
                ))
            else:
                edges[id(entry)].append((complement_txn, link.account))
                edges[id(complement_txn)].append((entry, link.complement_account))

        for complement_txn in filter_txns(complement_entries):
            complement_feature = _transaction_feature(
                complement_txn, link.complement_account, False)
            if complement_feature[1] and id(complement_txn) not in edges:
                logger.log_error(UnresolvedLinkError(
                    complement_txn.meta,
                    f'No complement transaction found for link {link}',
                    complement_txn,
                ))

    return edges


def _resolve_links(
        entries_by_file: dict[str, list[Directive]],
        edges: dict[int, list[tuple[Transaction, str]]],
        logger: error_lib.ErrorLogger,
) -> list[Directive]:
    ret = []
    all_visited = set()
    bad = set()
    for entries in entries_by_file.values():
        for entry in entries:
            if id(entry) in all_visited:
                continue
            if id(entry) not in edges or id(entry) in bad:
                ret.append(entry)
                continue

            q = collections.deque([entry])
            visited = set()
            txns = []
            while q:
                u = q.popleft()
                if id(u) in visited:
                    continue
                visited.add(id(u))
                txns.append(u)
                for edge in edges[id(u)]:
                    q.append(edge[0])
            merged_txn = merge_transactions(txns, edges, logger)
            if merged_txn:
                ret.append(merged_txn)
                all_visited.update(visited)
            else:
                ret.append(entry)
                bad.update(visited)
    return ret


def _transaction_feature(
        entry: Transaction,
        account: str,
        negated: bool,
) -> tuple[Optional[str], collections.Counter]:
    link_key = entry.meta.get('share_link_key', None)

    posting_features = []
    for posting in entry.postings:
        if _main_account(posting.account) != account:
            continue
        if negated:
            units = amount.mul(posting.units, decimal.Decimal(-1))
        else:
            units = posting.units
        posting_features.append(units)
    return link_key, collections.Counter(posting_features)


def merge_transactions(
        txns: list[Transaction],
        edges: dict[int, list[tuple[Transaction, str]]],
        logger: error_lib.ErrorLogger,
) -> Optional[Transaction]:
    date = None
    flag = None
    payee = None
    narration = ''
    tags = set[str]()
    links = set[str]()
    meta = dict[str, Any]()
    postings = []
    compatible = True
    last_txn = None
    for txn in txns:
        last_txn = txn
        if date and txn.date and txn.date != date:
            compatible = False
        date = date or txn.date
        if flag and txn.flag and txn.flag != flag:
            compatible = False
        flag = flag or txn.flag
        if payee and txn.payee and txn.payee != payee:
            compatible = False
        payee = payee or txn.payee
        if narration and txn.narration and txn.narration != narration:
            compatible = False
        narration = narration or txn.narration
        for k, v in txn.meta.items():
            mv = meta.get(k)
            meta[k] = mv or v
            if mv and mv != v and k not in ('filename', 'lineno'):
                compatible = False
                break
        if not compatible:
            break
        accounts_to_remove = set(edge[1] for edge in edges[id(txn)])
        for posting in txn.postings:
            if _main_account(posting.account) not in accounts_to_remove:
                postings.append(posting)

    if not compatible and last_txn:
        logger.log_error(UnresolvedLinkError(
            last_txn.meta,
            'Transaction and its complement do not agree on flag, payee, '
            'narration or meta',
            last_txn,
        ))
        return None
    txn = Transaction(
        date=date,
        meta=meta,
        flag=flag,
        payee=payee,
        narration=narration,
        tags=tags,
        links=links,
        postings=postings,
    )
    return txn



from typing import Any
from beancount.core.data import Directive, Transaction
from autobean.narration import comments


def plugin(entries: list[Directive], options: dict[str, Any]) -> tuple[list[Directive], list]:
    filenames = collect_transaction_filenames(entries)
    comment_narrations = {
        filename: comments.extract_from_file(filename)
        for filename in filenames
    }
    return [merge_narration(entry, comment_narrations) for entry in entries], []


def collect_transaction_filenames(entries: list[Directive]) -> set[str]:
    return {
        entry.meta['filename']
        for entry in entries
        if isinstance(entry, Transaction) and 'filename' in entry.meta
    }


def merge_narration(entry: Directive, comment_narrations: dict[str, dict[int, str]]) -> Directive:
    if not isinstance(entry, Transaction):
        return entry
    narrations = []
    for posting in entry.postings:
        if posting.meta:
            narration = posting.meta.get('narration')
            comment_narration = comment_narrations[posting.meta.get('filename')].get(posting.meta.get('lineno'))
            if narration is None and comment_narration:
                posting.meta['narration'] = comment_narration
                narration = comment_narration
            if narration:
                narrations.append(narration.strip())
    if entry.narration:
        return entry
    return entry._replace(narration=' | '.join(narrations))



import collections
import dataclasses
import io
import re
from typing import Any, Callable, Optional, Union
import os
import copy
import pytest
from beancount import loader
from beancount.core.data import Balance, Directive, Transaction, Open, Close
from beancount.parser import printer
from autobean.utils import error_lib
from autobean.utils.compare import compare_entries


NonParameterizedPlugin = Callable[[list[Directive], dict[str, Any]], tuple[list[Directive], list[error_lib.Error]]]
ParameterizedPlugin = Callable[[list[Directive], dict[str, Any], str], tuple[list[Directive], list[error_lib.Error]]]
Plugin = Union[NonParameterizedPlugin, ParameterizedPlugin]
_BRACKET_TO_ESCAPE_REGEX = re.compile(r':\[(.*?)\]')
_BRACKET_ESCAPED_REGEX = re.compile(r':TEST--(.*?)--$')


@dataclasses.dataclass(frozen=True)
class Ledger:
    entries: list[Directive]
    errors: list[error_lib.Error]
    options: dict[str, Any]


class ExpectedErrors:
    def __init__(self) -> None:
        self.errors = collections.defaultdict[tuple[str, int], list[str]](list)

    def add(self, filename: str, lineno: int, message: str) -> None:
        self.errors[(filename, lineno)].append(message)

    def format(self) -> str:
        return '\n'.join(
            f'{filename}:{lineno}:{message}'
            for (filename, lineno), messages in sorted(self.errors.items())
            for message in sorted(messages)
        )


@dataclasses.dataclass(frozen=True)
class Testcase:
    source: Ledger
    source_expected_errors: ExpectedErrors
    plugin_arg: Optional[str]
    expected_errors: ExpectedErrors
    expected_entries: Optional[list[Directive]]
    

def generate_tests(tests_path: str, plugin: Plugin) -> Callable[[Callable[[], None]], Callable[[Testcase], None]]:
    ids, args = collect_testcases(tests_path)
    def decorator(func: Callable[[], None]) -> Callable[[Testcase], None]:

        def test(testcase: Testcase) -> None:
            assert_same_errors(testcase.source.errors, testcase.source_expected_errors)
            entries, errors = apply_plugin(
                plugin, testcase.source.entries, testcase.source.options, testcase.plugin_arg)
            if testcase.expected_entries is not None:
                assert_same_results(entries, testcase.expected_entries)
            assert_same_errors(errors, testcase.expected_errors)
            func()

        return pytest.mark.parametrize('testcase', args, ids=ids)(test)
    return decorator


def collect_testcases(tests_path: str) -> tuple[list[str], list[Testcase]]:
    all_ids: list[str] = []
    all_testcases: list[Testcase] = []
    suites = os.listdir(tests_path)
    for suite in suites:
        ids, testcases = load_test_suite(tests_path, suite)
        all_ids += ids
        all_testcases += testcases
    return all_ids, all_testcases


def load_test_suite(tests_path: str, suite: str) -> tuple[list[str], list[Testcase]]:
    full_path = os.path.join(tests_path, suite)
    ids = list[str]()
    testcases = list[Testcase]()
    if not os.path.isdir(full_path) or suite.startswith('.') or suite.startswith('_'):
        return [], []
    names = set()
    for filename in os.listdir(full_path):
        segs = filename.rsplit('.', 2)
        if len(segs) != 2:
            continue
        names.add(segs[0])
    if 'source' not in names:
        return [], []
    names.remove('source')
    source, source_expected_errors = load_expected(full_path, 'source')
    assert source
    for name in names:
        if name.startswith('_'):
            continue
        plugin_arg = name if name != 'None' else None
        expected_ledger, expected_errors = load_expected(
            full_path, name, escape_brackets=True)
        ids.append(f'{suite} ({plugin_arg})')
        testcases.append(Testcase(
            source=copy.deepcopy(source),
            source_expected_errors=source_expected_errors,
            plugin_arg=plugin_arg,
            expected_entries=expected_ledger.entries if expected_ledger else None,
            expected_errors=expected_errors))
    return ids, testcases


def load_expected(
        path: str,
        name: str,
        *,
        escape_brackets: bool = False,
) -> tuple[Optional[Ledger], ExpectedErrors]:
    bean_path = os.path.join(path, f'{name}.bean')
    ledger = (
        load_ledger(bean_path, escape_brackets=escape_brackets)
        if os.path.isfile(bean_path) else None)
    errors_path = os.path.join(path, f'{name}.errors')
    errors = load_expected_errors(errors_path) if os.path.isfile(errors_path) else ExpectedErrors()
    return ledger, errors


def load_ledger(path: str, *, escape_brackets: bool = False) -> Ledger:
    if escape_brackets:
        with open(path) as f:
            content = f.read()
        content = _escape_output(content)
        entries, errors, options = loader.load_string(content)
        entries = _unescape_entries(entries)
    else:
        entries, errors, options = loader.load_file(path)
    return Ledger(entries, errors, options)


def load_expected_errors(path: str) -> ExpectedErrors:
    with open(path) as f:
        lines = f.readlines()
    errors = ExpectedErrors()
    for line in lines:
        filename, lineno, message = line.strip().split(':', 2)
        errors.add(filename, int(lineno), message)
    return errors


def assert_same_results(actuals: list[Directive], expecteds: list[Directive]) -> None:
    # string comparison for better error output
    actual_io = io.StringIO()
    printer.print_entries(actuals, file=actual_io)
    expected_io = io.StringIO()
    printer.print_entries(expecteds, file=expected_io)
    assert actual_io.getvalue() == expected_io.getvalue(), "unexpected output"

    # catch potential mismatch that isn't printed
    same, _, _= compare_entries(actuals, expecteds)
    assert same, "unexpected output"


def assert_same_errors(actuals: list[error_lib.Error], expecteds: ExpectedErrors) -> None:
    actual_errors = ExpectedErrors()
    matched_expected_errors = ExpectedErrors()
    unmatched_expected_errors = copy.deepcopy(expecteds)
    for actual in actuals:
        filename = os.path.basename(actual.source['filename'])
        lineno = actual.source['lineno']
        actual_errors.add(filename, lineno, actual.message)

        keywords = unmatched_expected_errors.errors.get((filename, lineno), [])
        for i, keyword in enumerate(keywords):
            if keyword in actual.message:
                matched_expected_errors.add(filename, lineno, actual.message)
                keywords.pop(i)
                break
    for (filename, lineno), keywords in unmatched_expected_errors.errors.items():
        for keyword in keywords:
            matched_expected_errors.add(filename, lineno, f'*{keyword}*')

    assert actual_errors.format() == matched_expected_errors.format(), "unexpected errors"


def apply_plugin(
        plugin: Callable,
        entries: list[Directive],
        options: dict[str, Any],
        plugin_arg: Optional[str]) -> tuple[list[Directive], list[error_lib.Error]]:

    if plugin_arg is None:
        return plugin(entries, options)
    return plugin(entries, options, plugin_arg)


def _escape_output(output: str) -> str:
    return re.sub(_BRACKET_TO_ESCAPE_REGEX, r':TEST--\1--', output)


def _unescape_account(account: str) -> str:
    """Unescapes square brackets from expected output.

    Expected output was escaped so it could parse.
    """
    return re.sub(_BRACKET_ESCAPED_REGEX, r':[\1]', account)


def _unescape_entries(entries: list[Directive]) -> list[Directive]:
    ret = []
    entries = sorted(entries, key=lambda e: e.meta['lineno'])
    for entry in entries:
        if isinstance(entry, Transaction):
            postings = [
                posting._replace(
                    account=_unescape_account(posting.account))
                for posting in entry.postings
            ]
            postings = sorted(postings, key=lambda p: p.meta['lineno'])
            ret.append(entry._replace(postings=postings))
        elif isinstance(entry, Open | Close | Balance):
            ret.append(entry._replace(
                account=_unescape_account(entry.account)))
        else:
            ret.append(entry)

    return ret


def generate_goldens(path: str, suite: str, plugin: Plugin) -> None:
    _, testcases = load_test_suite(path, suite)
    if not testcases:
        print(f'No testcases found for {suite}.')
    for testcase in testcases:
        entries, errors = apply_plugin(
            plugin, testcase.source.entries, testcase.source.options, testcase.plugin_arg)
        if testcase.expected_entries is not None:
            output_path = os.path.join(path, suite, f'{testcase.plugin_arg}.bean')
            with open(output_path, 'w') as f:
                printer.print_entries(entries, file=f)
            print(f'Generated {output_path}.')
        if errors:
            expected_errors = ExpectedErrors()
            for error in errors:
                expected_errors.add(
                    os.path.basename(error.source['filename']),
                    error.source['lineno'],
                    error.message)
            output_path = os.path.join(path, suite, f'{testcase.plugin_arg}.errors')
            with open(output_path, 'w') as f:
                f.write(expected_errors.format())
            print(f'Generated {output_path}.')



from collections import Counter
from beancount.core import data
from beancount.core.data import Directive
from beancount.core.compare import hash_entry


def compare_entries(entries1: list[Directive], entries2: list[Directive]) -> tuple[bool, list[Directive], list[Directive]]:
    """Compares two lists of entries.

    Similiar to beancount.core.comparer.compare_entries but allows duplicated entries
    """
    hashes1 = {
        hash_entry(entry, exclude_meta=True): entry for entry in entries1}
    hashes2 = {
        hash_entry(entry, exclude_meta=True): entry for entry in entries2}

    keys1 = Counter(hash_entry(entry, exclude_meta=True) for entry in entries1)
    keys2 = Counter(hash_entry(entry, exclude_meta=True) for entry in entries2)

    same = keys1 == keys2
    missing1 = data.sorted([hashes1[key] for key in keys1 - keys2])
    missing2 = data.sorted([hashes2[key] for key in keys2 - keys1])
    return same, missing1, missing2



import os.path
from typing import Any, Iterable, Optional
from beancount.core.data import Custom, Directive
from beancount import loader
from autobean.utils import plugin_lib


@plugin_lib.plugin('autobean.include')
class IncludePlugin(plugin_lib.BasePlugin):
    _includes: set[str]

    def process(self, entries: list[Directive], options: dict[str, Any], arg: Optional[str]) -> Iterable[Directive]:
        self._includes = set(options['include'])
        yield from super().process(entries, options, arg)
        # Allow tools to refresh data when included files are updated.
        options['include'] = list(self._includes)

    @plugin_lib.handle_custom('autobean.include', 'exactly one path')
    def _handle_include(self, custom: Custom, path: str) -> Iterable[Directive]:
        path = os.path.join(os.path.dirname(custom.meta['filename']), path)
        entries, errors, _ = loader.load_file(path)
        self._error_logger.log_loading_errors(errors, custom)
        self._includes.add(path)
        return entries



import datetime
from decimal import Decimal
import http.server
import logging
import os
import re
import time
import secrets
import sys
from typing import Any, Optional
import urllib.parse
import webbrowser

from autobean.utils import deduplicate
from beancount.core.amount import Amount
from beancount.core.data import Transaction, Posting, Balance, Directive, new_metadata
from beancount.core import inventory
from beancount.ingest import importer
from beancount.ingest import cache
import dateutil.parser
import requests
import yaml


CONFIG_SUFFIX = '.truelayer.yaml'
ACCOUNT_TYPES = ('accounts', 'cards')


def escape_account_component(s: str) -> str:
    s = re.sub(r'\W', '', s)
    s = s[:1].upper() + s[1:]
    return s


def format_iso_datetime(timestamp_s: float) -> str:
    return datetime.datetime.utcfromtimestamp(int(timestamp_s)).isoformat()


def currency_to_decimal(currency: float) -> Decimal:
    return Decimal(f'{currency:.2f}')


class Importer(importer.ImporterProtocol):

    def __init__(self, client_id: str, client_secret: str):
        self._client_id = client_id
        self._client_secret = client_secret

    def name(self) -> str:
        return 'autobean.truelayer'

    def identify(self, file: cache._FileMemo) -> bool:
        return file.name.endswith(CONFIG_SUFFIX)

    def extract(self, file: cache._FileMemo, existing_entries: Optional[list[Directive]] = None) -> list[Directive]:
        config = _Config(self._client_id, self._client_secret, file)
        extractor = _Extractor(config)
        return extractor.extract(existing_entries)


class _Config:
    def __init__(self, client_id: str, client_secret: str, file: cache._FileMemo):
        self.client_id = client_id
        self.client_secret = client_secret
        self.data = yaml.safe_load(file.contents()) or {}
        self._filename = file.name

    @property
    def name(self) -> str:
        return os.path.basename(self._filename).rsplit(CONFIG_SUFFIX, 1)[0]

    def dump(self) -> None:
        with open(self._filename, 'w') as f:
            yaml.safe_dump(self.data, f)


class _Extractor:
    def __init__(self, config: _Config):
        self._config = config
        self._oauth_manager = _OAuthManager(config)

    def extract(self, existing_entries: Optional[list[Directive]] = None) -> list[Directive]:
        for type_ in ACCOUNT_TYPES:
            self._update_accounts(type_)
        entries = self._fetch_all_transactions()
        if existing_entries:
            entries = deduplicate.deduplicate(entries, existing_entries)
        return entries

    @property
    def _auth_headers(self) -> dict[str, str]:
        return {
            'Authorization': f'Bearer {self._oauth_manager.access_token}'
        }

    def _update_accounts(self, type_: str) -> None:
        url = {
            'accounts': 'https://api.truelayer.com/data/v1/accounts',
            'cards': 'https://api.truelayer.com/data/v1/cards',
        }
        r = requests.get(url[type_], headers=self._auth_headers)
        if not r.ok:
            logging.warning('Could not fetch %s: %s', type_, r.text)
            return
        accounts = r.json().get('results', [])
        config_accounts = self._config.data.setdefault(type_, {})
        for account in accounts:
            config_account = config_accounts.setdefault(
                account['account_id'], {})
            config_account.setdefault('name', account['display_name'])
            config_account.setdefault(
                'liability',
                type_ == 'cards' and account['card_type'] == 'CREDIT')
            config_account.setdefault('enabled', True)
            config_account.setdefault('beancount_account', ':'.join([
                'Liabilities' if config_account['liability'] else 'Assets',
                escape_account_component(self._config.name),
                escape_account_component(config_account['name'])
            ]))
            config_account.setdefault('from', int(time.time()) - 86400 * 90)

        self._config.dump()

    def _fetch_transactions(
            self,
            account_id: str,
            account: dict[str, Any],
            type_: str,
            is_pending: bool) -> list[dict[str, Any]]:
        url = {
            ('accounts', False): (
                f'https://api.truelayer.com/data/v1/accounts/{account_id}/transactions'),
            ('accounts', True): (
                f'https://api.truelayer.com/data/v1/accounts/{account_id}/transactions/pending'),
            ('cards', False): (
                f'https://api.truelayer.com/data/v1/cards/{account_id}/transactions'),
            ('cards', True): (
                f'https://api.truelayer.com/data/v1/cards/{account_id}/transactions/pending'),
        }
        log_transaction = 'pending transactions' if is_pending else 'transactions'
        logging.info(
            f'Fetching {log_transaction} for account {account["name"]} '
            f'({account_id}).')
        r = requests.get(
            url[(type_, is_pending)],
            headers=self._auth_headers,
            params={
                'from': format_iso_datetime(account['from']),
                'to': format_iso_datetime(time.time()),
            }
        )
        if not r.ok:
            logging.error('Error fetching transactions: %s', r.text)
            r.raise_for_status()
        txns = r.json().get('results', [])
        logging.info(
            f'Fetched {len(txns)} {log_transaction} for account '
            f'{account["name"]} ({account_id}).')
        return txns

    def _fetch_balances(
            self,
            account_id: str,
            account: dict[str, Any],
            type_: str) -> list[dict[str, Any]]:
        url = {
            'accounts': f'https://api.truelayer.com/data/v1/accounts/{account_id}/balance',
            'cards': f'https://api.truelayer.com/data/v1/cards/{account_id}/balance',
        }
        logging.info(
            f'Fetching balance for account {account["name"]} ({account_id}).')
        r = requests.get(url[type_], headers=self._auth_headers)
        if not r.ok:
            logging.error('Error fetching balance: %s', r.text)
            r.raise_for_status()
        balances = r.json().get('results', [])
        logging.info(
            f'Fetched {len(balances)} balance entries for account '
            f'{account["name"]} ({account_id}).')
        return balances


    def _fetch_all_transactions(self) -> list[Directive]:
        entries: list[Directive] = []
        for type_ in ACCOUNT_TYPES:
            for account_id, account in self._config.data[type_].items():
                if not account['enabled']:
                    continue
                truelayer_txns = self._fetch_transactions(
                    account_id, account, type_, False)
                time_txns = [
                    (
                        dateutil.parser.parse(truelayer_txn['timestamp']),
                        self._transform_transaction(
                            truelayer_txn, account['beancount_account']))
                    for truelayer_txn in truelayer_txns
                ]
                pending_truelayer_txns = self._fetch_transactions(
                    account_id, account, type_, True)
                pending_time_txns = [
                    (
                        dateutil.parser.parse(truelayer_txn['timestamp']),
                        self._transform_transaction(
                            truelayer_txn, account['beancount_account'], True))
                    for truelayer_txn in pending_truelayer_txns
                ]
                entries.extend(txn for _, txn in time_txns)
                entries.extend(txn for _, txn in pending_time_txns)

                balances = self._fetch_balances(account_id, account, type_)
                for balance in balances:
                    entries.append(self._transform_balance(
                        balance, account, time_txns, pending_time_txns))
        return entries

    def _transform_balance(
            self,
            truelayer_balance: dict[str, Any],
            account: dict[str, Any],
            time_txns: list[tuple[datetime.datetime, Transaction]],
            pending_time_txns: list[tuple[datetime.datetime, Transaction]],
    ) -> Balance:
        """Transforms TrueLayer Balance to beancount Balance.
        
        Balance from TrueLayer can be effective at the middle of a day with
        pending transactions ignored, while beancount balance assertions
        must be applied at the beginning of a day and pending transactions
        are taken into account.

        It is not always possible to get pending transactions. If that is not
        available balance assertions may have to be corrected retrospectively.
        """

        balance_time = dateutil.parser.parse(
            truelayer_balance['update_timestamp']).astimezone()
        assertion_time = datetime.datetime.combine(
            balance_time, datetime.time.min, balance_time.tzinfo)

        txns_to_remove = [
            txn
            for t, txn in time_txns
            if assertion_time <= t < balance_time
        ]
        inventory_to_remove = inventory.Inventory()
        for txn in txns_to_remove:
            for posting in txn.postings:
                inventory_to_remove.add_position(posting)
        amount_to_remove = inventory_to_remove.get_currency_units(
            truelayer_balance['currency'])

        txns_to_add = [
            txn
            for t, txn in pending_time_txns
            if t < assertion_time
        ]
        inventory_to_add = inventory.Inventory()
        for txn in txns_to_add:
            for posting in txn.postings:
                inventory_to_add.add_position(posting)
        amount_to_add = inventory_to_add.get_currency_units(
            truelayer_balance['currency'])
        
        number = currency_to_decimal(truelayer_balance['current'])
        if account['liability']:
            number = -number
        number += amount_to_add.number
        number -= amount_to_remove.number
        return Balance(
            meta=new_metadata('', 0),
            date=assertion_time.date(),
            account=account['beancount_account'],
            amount=Amount(number, truelayer_balance['currency']),
            tolerance=None,
            diff_amount=None,
        )

    def _transform_transaction(
            self,
            truelayer_txn: dict[str, Any],
            beancount_account: str,
            is_pending: bool = False) -> Transaction:
        """Transforms TrueLayer Transaction to beancount Transaction."""

        number = abs(currency_to_decimal(truelayer_txn['amount']))
        if truelayer_txn['transaction_type'] == 'DEBIT':
            number = -number
        elif truelayer_txn['transaction_type'] == 'CREDIT':
            pass
        else:
            assert False

        posting = Posting(
            account=beancount_account,
            units=Amount(number, truelayer_txn['currency']),
            cost=None,
            price=None,
            flag=None,
            meta=None,
        )
        payee = (
            truelayer_txn.get('merchant_name', None) or
            truelayer_txn['meta'].get('provider_merchant_name', None))
        return Transaction(
            meta=new_metadata('', 0),
            date=dateutil.parser.parse(truelayer_txn['timestamp']).astimezone().date(),
            flag='!' if is_pending else '*',
            payee=payee,
            narration=truelayer_txn['description'],
            tags=set(),
            links=set(),
            postings=[posting],
        )


class _OAuthManager:
    ADDRESS = '127.0.0.1'
    PORT = 3000
    REDIRECT_URI = 'http://localhost:3000/callback'

    def __init__(self, config: _Config):
        self._config = config
    
    @property
    def access_token(self) -> str:
        access_token = self._get_valid_access_token()
        if access_token:
            return access_token
        self._refresh_access_token()
        access_token = self._get_valid_access_token()
        if access_token:
            return access_token
        self._request_access_token()
        access_token = self._get_valid_access_token()
        if access_token:
            return access_token
        raise RuntimeError('Unable to get a valid access token.')

    def _get_valid_access_token(self) -> Optional[str]:
        """Get access token from config file."""

        access_token = self._config.data.get('access_token')
        expiry_time = self._config.data.get('access_token_expiry_time')
        now = int(time.time())
        if access_token and expiry_time and expiry_time > now:
            return access_token
        return None

    def _refresh_access_token(self) -> None:
        """Refresh access token with refresh token."""

        logging.info('Attempt to refresh access token.')
        refresh_token = self._config.data.get('refresh_token', None)
        if not refresh_token:
            logging.info(
                'Failed to refresh access token: refresh token not available.')
            return
        self._grant_access_token(refresh_token=refresh_token)

    def _request_access_token(self) -> None:
        """Get access token with regular OAuth flow."""

        logging.info('Attempt to request access token with regular OAuth flow.')
        code = self._request_code()
        self._grant_access_token(code=code)
        logging.info('Successfully requested access token.')

    def _grant_access_token(self, code: Optional[str] = None, refresh_token: Optional[str] = None) -> None:
        """Grant access token with code or refresh_token."""

        logging.info('Attempt to grant access token.')
        req = {
            'client_id': self._config.client_id,
            'client_secret': self._config.client_secret,
        }
        if code:
            req['grant_type'] = 'authorization_code'
            req['redirect_uri'] = self.REDIRECT_URI
            req['code'] = code
        elif refresh_token:
            req['grant_type'] = 'refresh_token'
            req['refresh_token'] = refresh_token
        else:
            assert False
        r = requests.post('https://auth.truelayer.com/connect/token', req)
        if r.status_code != 200:
            logging.warning(
                f'Failed to grant access token: server returns '
                f'{r.status_code}')
            return
        data = r.json()
        self._config.data['access_token'] = data['access_token']
        self._config.data['access_token_expiry_time'] = (
            int(time.time()) + data['expires_in'])
        self._config.data['refresh_token'] = data['refresh_token']
        self._config.dump()
        logging.info('Successfully granted access token.')

    def _request_code(self) -> str:
        """Get the code to redeem access token with regular OAuth flow."""

        state = secrets.token_urlsafe(16)
        code = None
        auth_link = None

        class HttpHandler(http.server.BaseHTTPRequestHandler):
            def do_POST(self) -> None:
                logging.info('OAuth response received.')
                length = int(self.headers.get('Content-Length'))
                body = self.rfile.read(length).decode('utf-8')
                data = dict(urllib.parse.parse_qsl(body))
                received_state = data.get('state')
                received_code = data.get('code')

                if received_code and received_state == state:
                    nonlocal code
                    code = received_code
                    self.send_response(200)
                    response = b'You can now close this tab.\n'
                    self.send_header('Content-Type', 'text/plain')
                    self.send_header('Content-Length', str(len(response)))
                    self.end_headers()
                    self.wfile.write(response)
                else:
                    if received_state != state:
                        logging.warning('OAuth response state mismatches.')
                    elif not received_code:
                        logging.warning('OAuth response misses code.')
                    self.send_response(302)
                    assert auth_link
                    self.send_header('Location', auth_link)

        httpd = http.server.HTTPServer((self.ADDRESS, self.PORT), HttpHandler)
        socketname = httpd.socket.getsockname()
        logging.info(f'OAuth server listening at {socketname}')
        auth_link = self._build_auth_link(state)
        webbrowser.open_new(auth_link)

        print(
            f'Please navigate to the following URL to complete the '
            f'authorization process:\n\n'
            f'{auth_link}\n\n'
            f'If you are unable to visit the link on the same '
            f'host that this script is running on, you might need to '
            f'forward TCP port {socketname[1]} to the host where the '
            f'browser will be running on during the process.',
            file=sys.stderr)

        while not code:
            httpd.handle_request()
        httpd.server_close()
        return code

    def _build_auth_link(self, state: str) -> str:
        qs = urllib.parse.urlencode({
            'response_type': 'code',
            'response_mode': 'form_post',
            'client_id': self._config.client_id,
            'redirect_uri': self.REDIRECT_URI,
            'scope': ' '.join([
                'accounts',
                'cards',
                'transactions',
                'balance',
                'offline_access',
            ]),
            'state': state,
        })
        return f'https://auth.truelayer.com/?{qs}'

logging.basicConfig(level=os.getenv('LOG_LEVEL', 'WARNING'))



from typing import Any, Iterable, Optional
from collections import Counter
import os.path
import datetime
from beancount.core.data import Directive, Custom, Transaction, Balance, Posting
from beancount import loader
from beancount.ops.validation import ValidationError
from autobean.utils import error_lib, plugin_lib


class CrossCheckError(error_lib.Error):
    pass


class PostingToCompare:
    posting: Posting
    transaction: Transaction

    def __init__(self, posting: Posting, transaction: Transaction):
        self.posting = posting
        self.transaction = transaction

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, PostingToCompare) and
            self.transaction.date == other.transaction.date and
            self.posting.account == other.posting.account and
            self.posting.units == other.posting.units)

    def __hash__(self) -> int:
        return hash((self.transaction.date, self.posting.account, self.posting.units))


@plugin_lib.plugin('autobean.xcheck')
class CrossCheckPlugin(plugin_lib.BasePlugin):
    _includes: set[str]

    def process(self, entries: list[Directive], options: dict[str, Any], arg: Optional[str]) -> Iterable[Directive]:
        self._entries = entries
        self._includes = set(options['include'])
        yield from super().process(entries, options, arg)
        options['include'] = list(self._includes)

    @plugin_lib.handle_custom('autobean.xcheck', 'a path, a start date and zero or more accounts')
    def handle_xcheck(self, entry: Custom, path: str, start: datetime.date, *accounts_tuple: plugin_lib.Account) -> Iterable[Directive]:
        path = os.path.join(os.path.dirname(entry.meta['filename']), path)
        accounts = set[str](accounts_tuple)
        end = entry.date
        stmt_entries, stmt_errors, _ = loader.load_file(path)
        stmt_errors = [error for error in stmt_errors if not isinstance(error, ValidationError)]
        if stmt_errors:
            yield entry
            self._error_logger.log_loading_errors(stmt_errors, entry)
            return

        entries = _filter_by_time_period(self._entries, start, end)
        stmt_entries = _filter_by_time_period(stmt_entries, start, end)
        postings = list(_extract_related_postings(entries, accounts))
        stmt_postings = list(_extract_related_postings(stmt_entries, accounts))

        _, unexpected, missing = _compare_postings(postings, stmt_postings)
        for posting in unexpected:
            self._error_logger.log_error(CrossCheckError(
                posting.posting.meta, 'Unexpected posting', posting.transaction
            ))
        for posting in missing:
            self._error_logger.log_error(CrossCheckError(
                posting.posting.meta, 'Missing posting', posting.transaction
            ))
        self._includes.add(path)
        for stmt_entry in stmt_entries:
            if isinstance(stmt_entry, Balance):
                yield stmt_entry
        yield entry


def _extract_related_postings(entries: list[Directive], accounts: set[str]) -> Iterable[PostingToCompare]:
    for entry in entries:
        if not isinstance(entry, Transaction):
            continue
        for posting in entry.postings:
            if not accounts or posting.account in accounts:
                yield PostingToCompare(posting, entry)


def _compare_postings(postings1: list[PostingToCompare], postings2: list[PostingToCompare]) -> tuple[bool, list[PostingToCompare], list[PostingToCompare]]:
    missings2 = list(_find_missings(postings1, postings2))
    missings1 = list(_find_missings(postings2, postings1))
    same = not missings1 and not missings2
    return same, missings1, missings2


def _find_missings(postings1: Iterable[PostingToCompare], postings2: Iterable[PostingToCompare]) -> Iterable[PostingToCompare]:
    hashed1 = Counter(postings1)
    for posting in postings2:
        if hashed1[posting]:
            hashed1[posting] -= 1
        else:
            yield posting


def _filter_by_time_period(entries: Iterable[Directive], start: datetime.date, end: datetime.date) -> list[Directive]:
    return[entry for entry in entries if start <= entry.date < end]



import dataclasses
import decimal
import functools
from typing import Any, Generic, Optional, TypeVar
from beancount.core import account as beancount_account
from beancount.core.data import Balance, Custom, Posting
from autobean.utils import error_lib

_T = TypeVar('_T')
_U = TypeVar('_U')
_O = TypeVar('_O', bound='Ownership')


class _SpecialMeta:
    POLICY = 'share_policy'
    ENFORCED = 'share_enforced'
    CONVERSION = 'share_conversion'
    PRORATED = 'share_prorated'
    PRORATED_INCLUDED = 'share_prorated_included'
    ALL = frozenset([POLICY, ENFORCED, CONVERSION, PRORATED, PRORATED_INCLUDED])


class Ownership:
    pass


@dataclasses.dataclass(frozen=True)
class WeightedOwnership(Ownership):
    weights: dict[str, decimal.Decimal]

    @functools.cached_property
    def total_weight(self) -> decimal.Decimal:
        return sum(self.weights.values(), decimal.Decimal(0))


class ProratedOwnership(Ownership):
    pass


_PRORATED = ProratedOwnership()


@dataclasses.dataclass(frozen=True)
class Policy(Generic[_O]):
    ownership: _O
    enforced: bool
    conversion: bool
    prorated_included: bool


@dataclasses.dataclass(frozen=True)
class PolicyDefinition:
    parent: Optional[str]
    ownership: Optional[Ownership]
    enforced: Optional[bool]
    conversion: Optional[bool]
    prorated_included: Optional[bool]

    def as_policy(self) -> Optional[Policy]:
        if not self.ownership or self.enforced is None or self.conversion is None or self.prorated_included is None:
            return None
        return Policy(
            ownership=self.ownership,
            enforced=self.enforced,
            conversion=self.conversion,
            prorated_included=self.prorated_included,
        )


_ROOT_POLICY = PolicyDefinition(
    parent=None, ownership=None, enforced=False, conversion=True, prorated_included=True)


class PolicyLookupException(error_lib.PluginException):
    pass


def _check_type(key: str, value: Any, type_: type[_T]) -> _T:
    if not isinstance(value, type_):
        raise error_lib.PluginException(f'{key} must be a {type_.__name__}')
    return value


def _get_meta(meta: dict[str, Any], key: str, type_: type[_T], default: _U) -> _T | _U:
    value = meta.get(key)
    if value is None:
        return default
    return _check_type(key, value, type_)


def try_parse_policy_definition(meta: Optional[dict[str, Any]]) -> Optional[PolicyDefinition]:
    if not meta:
        return None
    weights = {}
    for key, value in meta.items():
        if key.startswith('share-'):
            name = key.removeprefix('share-')
            if not name or not name[0].isupper():
                raise error_lib.PluginException(
                    f'Name of share parties must be capitalized: got {name!r}')
            weights[name] = _check_type(key, value, decimal.Decimal)
        elif key.startswith('share_'):
            if key not in _SpecialMeta.ALL:
                raise error_lib.PluginException(
                    f'Unrecognized share_* metadata {key!r}')
    ownership: Optional[Ownership]
    if _get_meta(meta, _SpecialMeta.PRORATED, bool, None):
        if weights:
            raise error_lib.PluginException(
                f'Cannot use share-* metadata with share_prorated')
        ownership = _PRORATED
    elif weights:
        ownership = WeightedOwnership(weights)
    else:
        ownership = None

    parent = _get_meta(meta, _SpecialMeta.POLICY, str, None)
    balance = _get_meta(meta, _SpecialMeta.ENFORCED, bool, None)
    conversion = _get_meta(meta, _SpecialMeta.CONVERSION, bool, None)
    prorate_included = _get_meta(
        meta, _SpecialMeta.PRORATED_INCLUDED, bool, None)
    if not ownership and parent is None and balance is None and conversion is None and prorate_included is None:
        return None
    return PolicyDefinition(
        parent=parent,
        ownership=ownership,
        enforced=balance,
        conversion=conversion,
        prorated_included=prorate_included,
    )


def strip_share_meta(meta: Optional[dict[str, Any]]) -> None:
    if not meta:
        return
    for key in list(meta):
        if key.startswith('share-') or key.startswith('share_'):
            del meta[key]


class PolicyDatabase:
    _named_policies: dict[str, PolicyDefinition]
    _account_policies: dict[str, PolicyDefinition]
    _wildcard_account_policies: dict[str, PolicyDefinition]

    def __init__(self) -> None:
        self._named_policies = {}
        self._account_policies = {}
        self._wildcard_account_policies = {}

    def add_policy(self, name: str, policy_def: PolicyDefinition) -> None:
        if policy_def.parent and policy_def.parent not in self._named_policies:
            raise error_lib.PluginException(
                f'Reference to unknown share policy {policy_def.parent!r}')
        if policy_def.enforced and not policy_def.ownership:
            raise error_lib.PluginException(
                f'Policy with share_enforced must define ownership')
        if name.endswith(':*'):
            self._wildcard_account_policies[name.removesuffix(':*')] = policy_def
        elif ':' in name:
            self._account_policies[name] = policy_def
        else:
            self._named_policies[name] = policy_def

    def _get_account_policy_definition(self, account: str) -> Optional[PolicyDefinition]:
        policy_def = None
        for parent in reversed(list(beancount_account.parents(account))):
            if wildcard_policy_def := self._wildcard_account_policies.get(parent):
                policy_def = _override_policy_def(
                    self._resolve_parent(wildcard_policy_def), policy_def, is_ephemeral=False)
        if account_policy_def := self._account_policies.get(account):
            policy_def = _override_policy_def(
                self._resolve_parent(account_policy_def), policy_def, is_ephemeral=False)
        return policy_def

    def _resolve_parent(self, policy_def: PolicyDefinition) -> PolicyDefinition:
        # TODO: maybe add loop detection?
        while policy_def.parent:
            policy_def = _override_policy_def(
                policy_def, self._named_policies[policy_def.parent], is_ephemeral=False)
        return policy_def

    def get_posting_policy(
            self,
            posting: Posting,
            transaction_policy_def: Optional[PolicyDefinition],
    ) -> Policy:
        policy_def = _ROOT_POLICY
        if default_policy_def := self._named_policies.get('default'):
            policy_def = _override_policy_def(
                self._resolve_parent(default_policy_def), policy_def, is_ephemeral=False)
        if transaction_policy_def:
            policy_def = _override_policy_def(
                self._resolve_parent(transaction_policy_def), policy_def, is_ephemeral=True)
        if account_policy_def := self._get_account_policy_definition(posting.account):
            policy_def = _override_policy_def(
                account_policy_def, policy_def, is_ephemeral=False)
        if posting_policy := try_parse_policy_definition(posting.meta):
            policy_def = _override_policy_def(
                self._resolve_parent(posting_policy), policy_def, is_ephemeral=True)
        if policy := policy_def.as_policy():
            return policy
        raise error_lib.PluginException('No applicable share policy')

    def get_balance_policy(self, balance: Balance) -> Optional[Policy[WeightedOwnership]]:
        has_explicit_policy = False
        balance_policy_def = try_parse_policy_definition(balance.meta)
        if balance_policy_def:
            balance_policy_def = self._resolve_parent(balance_policy_def)
        if balance_policy_def and balance_policy_def.ownership:
            if not isinstance(balance_policy_def.ownership, WeightedOwnership):
                raise error_lib.PluginException(
                    'Share policy on balance directive must have weighted ownership')
            has_explicit_policy = True
        policy_def = _ROOT_POLICY
        if account_policy_def := self._get_account_policy_definition(balance.account):
            policy_def = _override_policy_def(
                account_policy_def, policy_def, is_ephemeral=False)
        if balance_policy_def:
            policy_def = _override_policy_def(
                balance_policy_def, policy_def, is_ephemeral=True)
        if isinstance(policy_def.ownership, WeightedOwnership) and (policy_def.enforced or has_explicit_policy):
            return policy_def.as_policy()
        return None

    def get_proportionate_policy(self, custom: Custom, account: str) -> Policy[WeightedOwnership]:
        policy_def = _ROOT_POLICY
        if default_policy_def := self._named_policies.get('default'):
            policy_def = _override_policy_def(
                self._resolve_parent(default_policy_def), policy_def, is_ephemeral=False)
        if account_policy_def := self._get_account_policy_definition(account):
            policy_def = _override_policy_def(
                account_policy_def, policy_def, is_ephemeral=False)
        if custom_policy_def := try_parse_policy_definition(custom.meta):
            policy_def = _override_policy_def(
                self._resolve_parent(custom_policy_def), policy_def, is_ephemeral=True)
        policy = policy_def.as_policy()
        if not policy:
            raise error_lib.PluginException('No applicable share policy')
        if not isinstance(policy.ownership, WeightedOwnership):
            raise error_lib.PluginException(
                'Share policy on autobean.share.proportionate must have weighted ownership')
        return policy


def _override_policy_def(
        policy_def: PolicyDefinition,
        parent: Optional[PolicyDefinition],
        *,
        is_ephemeral: bool,
) -> PolicyDefinition:
    if parent is None:
        return policy_def
    if is_ephemeral:
        if parent.enforced and policy_def.ownership:
            raise error_lib.PluginException(
                'Cannot temporarily override ownership of a policy with share_enforced: TRUE')
        if parent.enforced and policy_def.enforced == False:
            raise error_lib.PluginException(
                'Cannot unset share_enforced')
    ownership = policy_def.ownership or parent.ownership
    balance = policy_def.enforced if policy_def.enforced is not None else parent.enforced
    conversion = policy_def.conversion if policy_def.conversion is not None else parent.conversion
    prorate_included = policy_def.prorated_included if policy_def.prorated_included is not None else parent.prorated_included
    return PolicyDefinition(
        parent=parent.parent,
        ownership=ownership,
        enforced=balance,
        conversion=conversion,
        prorated_included=prorate_included,
    )



import decimal
import textwrap
from typing import Any, Callable, Iterator, Optional, TypeVar
from beancount.parser import parser
from beancount.core.data import Balance, Custom, Directive, Transaction
import pytest
from autobean.utils import error_lib
from . import policy_lib

_T = TypeVar('_T', bound=Directive)
_U = TypeVar('_U')


def _parse_doc(type: _T) -> Callable[[Callable[..., _U]], Callable[..., _U]]:
    def decorator(func: Callable[..., _U]) -> Callable[..., _U]:
        assert func.__doc__
        directive = parser.parse_one(textwrap.dedent(func.__doc__))
        assert isinstance(directive, type)
        def wrapped(*args: Any, **kwargs: Any) -> _U:
            return func(*args, directive, **kwargs)
        return wrapped
    return decorator


@_parse_doc(Transaction)
def test_parse_weighted(txn: Transaction) -> None:
    """
    2000-01-01 *
        share-Alice: 1
        share-Bob: 1.5
    """
    policy_def = policy_lib.try_parse_policy_definition(txn.meta)
    assert policy_def
    assert isinstance(policy_def.ownership, policy_lib.WeightedOwnership)
    assert policy_def.ownership.total_weight == 2.5
    assert policy_def.ownership.weights == {
        'Alice': 1,
        'Bob': 1.5,
    }
    assert policy_def.parent is None
    assert policy_def.enforced is None
    assert policy_def.conversion is None
    assert policy_def.prorated_included is None


@_parse_doc(Transaction)
def test_parse_options_prorated(txn: Transaction) -> None:
    """
    2000-01-01 *
        share_prorated: TRUE
        Assets:Foo  100.00 USD
        Assets:Bar -100.00 USD
    """
    policy_def = policy_lib.try_parse_policy_definition(txn.meta)
    assert policy_def
    assert isinstance(policy_def.ownership, policy_lib.ProratedOwnership)


@_parse_doc(Transaction)
def test_parse_empty(txn: Transaction) -> None:
    """
    2000-01-01 *
        something: 123
        Assets:Foo  100.00 USD
        Assets:Bar -100.00 USD
    """
    policy_def = policy_lib.try_parse_policy_definition(txn.meta)
    assert policy_def is None


@_parse_doc(Transaction)
def test_parse_options_only(txn: Transaction) -> None:
    """
    2000-01-01 *
        something: 123
        share_conversion: FALSE
    """
    policy_def = policy_lib.try_parse_policy_definition(txn.meta)
    assert policy_def
    assert policy_def.ownership is None
    assert policy_def.conversion == False
    assert policy_def.prorated_included is None


@_parse_doc(Transaction)
def test_parse_bad_name(txn: Transaction) -> None:
    """
    2000-01-01 *
        share-alice: 1
    """
    with pytest.raises(error_lib.PluginException, match='capitalized.*alice'):
        policy_lib.try_parse_policy_definition(txn.meta)


@_parse_doc(Transaction)
def test_parse_options_conflicted_ownership(txn: Transaction) -> None:
    """
    2000-01-01 *
        share_prorated: TRUE
        share-Alice: 1
        share-Bob: 1
        something: 123
    """
    with pytest.raises(error_lib.PluginException, match='with share_prorated'):
        policy_lib.try_parse_policy_definition(txn.meta)


@_parse_doc(Transaction)
def test_parse_unrecognized_option(txn: Transaction) -> None:
    """
    2000-01-01 *
        share-Alice: 1
        share-Bob: 1
        share_foo: 1
    """
    with pytest.raises(error_lib.PluginException, match='Unrecognized.*foo'):
        policy_lib.try_parse_policy_definition(txn.meta)


class _TestPolicyDatabase:

    @pytest.fixture(autouse=True)
    def _setup_policy_db(self) -> Iterator[None]:
        self._policy_db = policy_lib.PolicyDatabase()
        yield None

    def _create_policy(self, id: Optional[int] = None, **options: Any) -> policy_lib.PolicyDefinition:
        ownership: Optional[policy_lib.Ownership]
        if id is not None:
            ownership = policy_lib.WeightedOwnership(weights={'Alice': decimal.Decimal(id)})
        elif options.pop('prorated', None):
            ownership = policy_lib._PRORATED
        else:
            ownership = None
        return policy_lib.PolicyDefinition(
            ownership=ownership,
            **{
                'parent': None,
                'enforced': None,
                'conversion': None,
                'prorated_included': None,
                **options,
            })


class TestAddPolicy(_TestPolicyDatabase):

    def test_add_policy_unknown_parent(self) -> None:
        policy = self._create_policy(1, parent='foo')
        with pytest.raises(error_lib.PluginException, match='unknown.*foo'):
            self._policy_db.add_policy('bar', policy)

    def test_add_balance_policy_without_ownership(self) -> None:
        policy = self._create_policy(enforced=True)
        with pytest.raises(error_lib.PluginException, match='share_enforced.*ownership'):
            self._policy_db.add_policy('foo', policy)


class TestPostingPolicyResolution(_TestPolicyDatabase):

    @_parse_doc(Transaction)
    def test_keep_going(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            share-Alice: 1
            Assets:Account 100.00 USD
                share-Alice: 42
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(2, conversion=False))
        self._policy_db.add_policy('Assets:Account:*', self._create_policy(3, prorated_included=True))
        self._policy_db.add_policy('Assets:*', self._create_policy(4, prorated_included=False))
        self._policy_db.add_policy('default', self._create_policy(5, conversion=True))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert policy.conversion == False
        assert policy.prorated_included == True

    @_parse_doc(Transaction)
    def test_priority_posting(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            share-Alice: 1
            Assets:Account 100.00 USD
                share-Alice: 42
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(2))
        self._policy_db.add_policy('Assets:Account:*', self._create_policy(3))
        self._policy_db.add_policy('Assets:*', self._create_policy(4))
        self._policy_db.add_policy('default', self._create_policy(5))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Transaction)
    def test_priority_account(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            share-Alice: 1
            Assets:Account 100.00 USD
                share_prorated_included: FALSE
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(42))
        self._policy_db.add_policy('Assets:Account:*', self._create_policy(2))
        self._policy_db.add_policy('Assets:*', self._create_policy(3))
        self._policy_db.add_policy('default', self._create_policy(4))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Transaction)
    def test_priority_wildcard_account_same(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            share-Alice: 1
            Assets:Account 100.00 USD
                share_prorated_included: FALSE
        """
        self._policy_db.add_policy('Assets:Account:*', self._create_policy(42))
        self._policy_db.add_policy('Assets:*', self._create_policy(2))
        self._policy_db.add_policy('default', self._create_policy(3))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Transaction)
    def test_priority_wildcard_account_parent(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            share-Alice: 1
            Assets:Account 100.00 USD
                share_prorated_included: FALSE
        """
        self._policy_db.add_policy('Assets:*', self._create_policy(42))
        self._policy_db.add_policy('default', self._create_policy(2))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Transaction)
    def test_priority_transaction(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            share-Alice: 42
            Assets:Account 100.00 USD
                share_prorated_included: FALSE
        """
        self._policy_db.add_policy('default', self._create_policy(1))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Transaction)
    def test_priority_default(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
                share_prorated_included: FALSE
        """
        self._policy_db.add_policy('default', self._create_policy(42))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Transaction)
    def test_missing(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
                share_conversion: FALSE
        """
        self._policy_db.add_policy('Liabilities:*', self._create_policy(1))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        with pytest.raises(error_lib.PluginException, match='No applicable share policy'):
            self._policy_db.get_posting_policy(txn.postings[0], policy_def)

    @_parse_doc(Transaction)
    def test_incomplete(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            share_conversion: FALSE
            Assets:Account 100.00 USD
                share_conversion: FALSE
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(conversion=True))
        self._policy_db.add_policy('Assets:Account:*', self._create_policy(conversion=False))
        self._policy_db.add_policy('Assets:*', self._create_policy(conversion=True))
        self._policy_db.add_policy('default', self._create_policy(conversion=True))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        with pytest.raises(error_lib.PluginException, match='No applicable share policy'):
            self._policy_db.get_posting_policy(txn.postings[0], policy_def)

    @_parse_doc(Transaction)
    def test_balance_policy_ownership_override_rejected(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
                share-Bob: 1
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(42, enforced=True))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        with pytest.raises(error_lib.PluginException, match='override.*share_enforced'):
            self._policy_db.get_posting_policy(txn.postings[0], policy_def)

    @_parse_doc(Transaction)
    def test_share_enforced_ephemeral_unset_rejected(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
                share_enforced: FALSE
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(42, enforced=True))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        with pytest.raises(error_lib.PluginException, match='unset.*share_enforced'):
            self._policy_db.get_posting_policy(txn.postings[0], policy_def)

    @_parse_doc(Transaction)
    def test_share_enforced_unset_accepted(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
                share_enforced: FALSE
        """
        self._policy_db.add_policy('Assets:*', self._create_policy(1, enforced=True))
        self._policy_db.add_policy('Assets:Account', self._create_policy(42, enforced=False))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42
        assert policy.enforced == False

    @_parse_doc(Transaction)
    def test_balance_policy_option_accepted(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
                share_conversion: FALSE
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(
            42, enforced=True, conversion=True))

        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert policy.conversion == False

    @_parse_doc(Transaction)
    def test_prorated(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            share_prorated: TRUE
            Assets:Account 100.00 USD
        """
        policy_def = policy_lib.try_parse_policy_definition(txn.meta)
        policy = self._policy_db.get_posting_policy(txn.postings[0], policy_def)
        assert isinstance(policy.ownership, policy_lib.ProratedOwnership)

    @_parse_doc(Transaction)
    def test_update_policy(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(1))
        self._policy_db.add_policy('Assets:Account', self._create_policy(42))

        policy = self._policy_db.get_posting_policy(txn.postings[0], None)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Transaction)
    def test_move_pointer(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
                share_policy: "foo"
        """
        self._policy_db.add_policy('aaa', self._create_policy(1))
        self._policy_db.add_policy('bbb', self._create_policy(42))
        self._policy_db.add_policy('foo', self._create_policy(parent='aaa'))
        self._policy_db.add_policy('foo', self._create_policy(parent='bbb'))

        policy = self._policy_db.get_posting_policy(txn.postings[0], None)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Transaction)
    def test_move_pointer_indirect(self, txn: Transaction) -> None:
        """
        2000-01-01 *
            Assets:Account 100.00 USD
                share_policy: "foo"
        """
        self._policy_db.add_policy('aaa', self._create_policy(1))
        self._policy_db.add_policy('bbb', self._create_policy(42))
        self._policy_db.add_policy('bar', self._create_policy(parent='aaa'))
        self._policy_db.add_policy('foo', self._create_policy(parent='bar'))
        self._policy_db.add_policy('bar', self._create_policy(parent='bbb'))

        policy = self._policy_db.get_posting_policy(txn.postings[0], None)
        assert isinstance(policy.ownership, policy_lib.WeightedOwnership)
        assert policy.ownership.weights['Alice'] == 42


class TestBalancePolicyResolution(_TestPolicyDatabase):

    @_parse_doc(Balance)
    def test_priority_explicit(self, balance: Balance) -> None:
        """
        2000-01-01 balance Assets:Account 100.00 USD
            share-Alice: 42
            share_conversion: FALSE
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(1, conversion=True))
        self._policy_db.add_policy('Assets:*', self._create_policy(conversion=True))
        self._policy_db.add_policy('default', self._create_policy(conversion=True))

        policy = self._policy_db.get_balance_policy(balance)
        assert policy
        assert policy.enforced == False
        assert policy.conversion == False
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Balance)
    def test_priority_account(self, balance: Balance) -> None:
        """
        2000-01-01 balance Assets:Account 100.00 USD
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(conversion=False))
        self._policy_db.add_policy('Assets:*', self._create_policy(42, conversion=True, enforced=True))
        self._policy_db.add_policy('default', self._create_policy(conversion=True))

        policy = self._policy_db.get_balance_policy(balance)
        assert policy
        assert policy.enforced == True
        assert policy.conversion == False
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Balance)
    def test_priority_wildcard_account(self, balance: Balance) -> None:
        """
        2000-01-01 balance Assets:Account 100.00 USD
        """
        self._policy_db.add_policy('Assets:*', self._create_policy(42, conversion=False, enforced=True))
        self._policy_db.add_policy('default', self._create_policy(conversion=True))

        policy = self._policy_db.get_balance_policy(balance)
        assert policy
        assert policy.enforced == True
        assert policy.conversion == False
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Balance)
    def test_non_balance_policy_ignored(self, balance: Balance) -> None:
        """
        2000-01-01 balance Assets:Account 100.00 USD
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(42))

        policy = self._policy_db.get_balance_policy(balance)
        assert not policy

    @_parse_doc(Balance)
    def test_default_ignored(self, balance: Balance) -> None:
        """
        2000-01-01 balance Assets:Account 100.00 USD
        """
        self._policy_db.add_policy('default', self._create_policy(42, conversion=True))

        policy = self._policy_db.get_balance_policy(balance)
        assert not policy

    @_parse_doc(Balance)
    def test_non_weighted_explicit(self, balance: Balance) -> None:
        """
        2000-01-01 balance Assets:Account 100.00 USD
            share_prorated: TRUE
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(42))

        with pytest.raises(error_lib.PluginException, match='must.*weighted'):
            self._policy_db.get_balance_policy(balance)

    @_parse_doc(Balance)
    def test_non_weighted_implicit(self, balance: Balance) -> None:
        """
        2000-01-01 balance Assets:Account 100.00 USD
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(prorated=True))

        policy = self._policy_db.get_balance_policy(balance)
        assert not policy

    @_parse_doc(Balance)
    def test_missing_implicit(self, balance: Balance) -> None:
        """
        2000-01-01 balance Assets:Account 100.00 USD
        """
        policy = self._policy_db.get_balance_policy(balance)
        assert not policy


class TestProportionatePolicyResolution(_TestPolicyDatabase):

    @_parse_doc(Custom)
    def test_priority_explicit(self, custom: Custom) -> None:
        """
        2000-01-01 custom "autobean.share.proportionate" Assets:Account
            share-Alice: 42
            share_conversion: FALSE
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(1, conversion=True))
        self._policy_db.add_policy('Assets:*', self._create_policy(conversion=True))
        self._policy_db.add_policy('default', self._create_policy(conversion=True))

        policy = self._policy_db.get_proportionate_policy(custom, 'Assets:Account')
        assert policy
        assert policy.conversion == False
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Custom)
    def test_priority_account(self, custom: Custom) -> None:
        """
        2000-01-01 custom "autobean.share.proportionate" Assets:Account
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(42, conversion=False))
        self._policy_db.add_policy('Assets:*', self._create_policy(1, conversion=True))
        self._policy_db.add_policy('default', self._create_policy(conversion=True))

        policy = self._policy_db.get_proportionate_policy(custom, 'Assets:Account')
        assert policy
        assert policy.conversion == False
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Custom)
    def test_priority_wildcard_account(self, custom: Custom) -> None:
        """
        2000-01-01 custom "autobean.share.proportionate" Assets:Account
        """
        self._policy_db.add_policy('Assets:*', self._create_policy(42, conversion=False))
        self._policy_db.add_policy('default', self._create_policy(1, conversion=True))

        policy = self._policy_db.get_proportionate_policy(custom, 'Assets:Account')
        assert policy
        assert policy.conversion == False
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Custom)
    def test_priority_default(self, custom: Custom) -> None:
        """
        2000-01-01 custom "autobean.share.proportionate" Assets:Account
        """
        self._policy_db.add_policy('default', self._create_policy(42, conversion=False))

        policy = self._policy_db.get_proportionate_policy(custom, 'Assets:Account')
        assert policy
        assert policy.conversion == False
        assert policy.ownership.weights['Alice'] == 42

    @_parse_doc(Custom)
    def test_non_weighted(self, custom: Custom) -> None:
        """
        2000-01-01 custom "autobean.share.proportionate" Assets:Account
        """
        self._policy_db.add_policy('Assets:Account', self._create_policy(prorated=True))

        with pytest.raises(error_lib.PluginException, match='must.*weighted'):
            self._policy_db.get_proportionate_policy(custom, 'Assets:Account')

    @_parse_doc(Custom)
    def test_missing(self, custom: Custom) -> None:
        """
        2000-01-01 custom "autobean.share.proportionate" Assets:Account
        """
        with pytest.raises(error_lib.PluginException, match='No applicable'):
            self._policy_db.get_proportionate_policy(custom, 'Assets:Account')



from typing import Any, Iterable, Optional
import os.path
from beancount.core.data import Custom, Directive, Open, Close, entry_sortkey
from beancount import loader
from autobean.utils import error_lib, plugin_lib
from . import link_accounts


@plugin_lib.plugin('autobean.share.include')
class IncludePlugin(plugin_lib.BasePlugin):

    def process(self, entries: list[Directive], options: dict[str, Any], arg: Optional[str]) -> Iterable[Directive]:
        self._enabled = True
        self._includes = set(options['include'])
        self._entries_by_file = dict[str, list[Directive]]()
        self._links = list[link_accounts.Link]()
        entries = list(super().process(entries, options, arg))
        entries += link_accounts.link_accounts(
            self._entries_by_file, self._links, self._error_logger)
        entries = deduplicate_open_close(entries)
        options['include'] = list(self._includes)
        return entries

    def _check_enabled(self, entry: Custom) -> None:
        if not self._enabled:
            raise error_lib.PluginException(
                f'Cannot use {entry.type} while autobean.share is disabled')

    @plugin_lib.handle_custom('autobean.share.enable', 'exactly one bool')
    def handle_enable(self, entry: Custom, enable: bool) -> Iterable[Directive]:
        self._enabled = enable
        yield entry

    @plugin_lib.handle_custom('autobean.share.include', 'exactly one path')
    def handle_include(self, entry: Custom, path: str) -> Iterable[Directive]:
        self._check_enabled(entry)
        path = os.path.join(os.path.dirname(entry.meta['filename']), path)
        if path in self._entries_by_file:
            return ()
        entries, errors, options = loader.load_file(path)
        self._error_logger.log_errors(errors)
        self._includes.update(options['include'])
        self._entries_by_file[path] = entries
        return ()

    @plugin_lib.handle_custom('autobean.share.link', 'path to an included ledger, an account in it, path to another included ledger, and an account in it')
    def handle_link(self, entry: Custom, path: str, account: plugin_lib.Account, complement_path: str, complement_account: plugin_lib.Account) -> Iterable[Directive]:
        self._check_enabled(entry)
        self._links.append(link_accounts.Link(
            path=path,
            account=account,
            complement_path=complement_path,
            complement_account=complement_account,
            custom=entry,
        ))
        return ()


def deduplicate_open_close(entries: list[Directive]) -> list[Directive]:
    """Deduplicates Open / Close directives.
    
    This deduplicates Open / Close directives in a simple way because duplication inside each file should have been
    caught during loading.
    """
    results = []
    opened_accounts = set()
    closed_accounts = set()
    for entry in sorted(entries, key=entry_sortkey):
        if isinstance(entry, Open):
            if entry.account not in opened_accounts:
                opened_accounts.add(entry.account)
                results.append(entry)
        elif isinstance(entry, Close):
            if entry.account not in closed_accounts:
                closed_accounts.add(entry.account)
                results.append(entry)
        else:
            results.append(entry)
    return results



from typing import Dict, Optional


def extract_from_file(filename: str) -> dict[int, str]:
    ret = {}
    with open(filename) as f:
        for i, line in enumerate(f):
            narration = extract_from_line(line)
            if narration is not None:
                ret[i + 1] = narration
    return ret


def extract_from_line(s: str) -> Optional[str]:
    segs = s.strip().split(';')
    return segs[2].strip() if len(segs) > 2 and not segs[1] else None



import datetime
import io
import re
import textwrap
from typing import Any
from beancount.parser import printer
from beancount.core.data import Directive, Transaction
from beancount.ops import balance
from beancount.parser import booking, parser
import pytest
from .plugin import Plugin


_FOO_TEXT = textwrap.dedent('''
    2000-01-01 open Assets:Foo
    2000-01-01 open Assets:Bar
    2000-01-01 open Income:Foo
    2000-02-01 *
        Income:Foo   -500.00 USD
        Assets:Foo   100.00 STOCK {{500 USD}}

    2000-03-01 *
        Income:Foo
        Assets:Foo   100.00 STOCK {6 USD}

    2000-04-01 *
        Income:Foo
        Assets:Bar   100.00 STOCK {6 USD}
    
    2000-05-01 balance Assets:Foo 200.00 STOCK
    2000-05-01 balance Assets:Bar 100.00 STOCK
    
    2000-05-01 custom "autobean.stock_split" 10 STOCK

    2000-05-02 balance Assets:Foo 2000.00 STOCK
    2000-05-02 balance Assets:Bar 1000.00 STOCK

    2000-05-03 *
        Income:Foo   -1 USD
        Assets:Bar   1 STOCK.B {{1 USD}}

    2000-05-04 balance Assets:Foo 2000.00 STOCK
    2000-05-04 balance Assets:Bar 1000.00 STOCK
    2000-05-04 balance Assets:Bar 1 STOCK.B

    2000-05-05 custom "autobean.stock_split" 2 STOCK

    2000-05-06 balance Assets:Foo 4000.00 STOCK
    2000-05-06 balance Assets:Bar 2000.00 STOCK
    2000-05-06 balance Assets:Bar 1 STOCK.B

    2000-05-07 custom "autobean.stock_split" 2 STOCK.B

    2000-05-08 balance Assets:Foo 4000.00 STOCK
    2000-05-08 balance Assets:Bar 2000.00 STOCK
    2000-05-08 balance Assets:Bar 2 STOCK.B
''')


def load(text: str) -> tuple[list[Directive], list[Any]]:
    entries, parsing_errors, options_map = parser.parse_string(text)
    entries, booking_errors = booking.book(entries, options_map)
    entries, plugin_errors = Plugin.plugin(entries, options_map)
    entries, balance_errors = balance.check(entries, options_map)
    return entries, [
        *parsing_errors,
        *booking_errors,
        *plugin_errors,
        *balance_errors,
    ]


def test_ok() -> None:
    entries, errors = load(_FOO_TEXT)
    assert not errors
    txn = entries[-15]
    assert isinstance(txn, Transaction)
    assert txn.date == datetime.date(2000, 5, 1)
    assert txn.narration == 'STOCK split 10:1'
    f = io.StringIO()
    printer.print_entry(txn, file=f)
    text = '\n'.join(sorted(filter(None, re.sub(r' +', ' ', f.getvalue()).split('\n')[1:])))
    assert text == '''\
 Assets:Bar -100.00 STOCK {6 USD, 2000-04-01}
 Assets:Bar 1000.00 STOCK {0.6 USD, 2000-04-01}
 Assets:Foo -100.00 STOCK {5 USD, 2000-02-01}
 Assets:Foo -100.00 STOCK {6 USD, 2000-03-01}
 Assets:Foo 1000.00 STOCK {0.5 USD, 2000-02-01}
 Assets:Foo 1000.00 STOCK {0.6 USD, 2000-03-01}\
'''

    txn = entries[-8]
    assert isinstance(txn, Transaction)
    assert txn.date == datetime.date(2000, 5, 5)
    assert txn.narration == 'STOCK split 2:1'
    f = io.StringIO()
    printer.print_entry(txn, file=f)
    text = '\n'.join(sorted(filter(None, re.sub(r' +', ' ', f.getvalue()).split('\n')[1:])))
    assert text == '''\
 Assets:Bar -1000.00 STOCK {0.6 USD, 2000-04-01}
 Assets:Bar 2000.00 STOCK {0.3 USD, 2000-04-01}
 Assets:Foo -1000.00 STOCK {0.5 USD, 2000-02-01}
 Assets:Foo -1000.00 STOCK {0.6 USD, 2000-03-01}
 Assets:Foo 2000.00 STOCK {0.25 USD, 2000-02-01}
 Assets:Foo 2000.00 STOCK {0.3 USD, 2000-03-01}\
'''

    txn = entries[-4]
    assert isinstance(txn, Transaction)
    assert txn.date == datetime.date(2000, 5, 7)
    assert txn.narration == 'STOCK.B split 2:1'
    f = io.StringIO()
    printer.print_entry(txn, file=f)
    text = '\n'.join(sorted(filter(None, re.sub(r' +', ' ', f.getvalue()).split('\n')[1:])))
    assert text == '''\
 Assets:Bar -1 STOCK.B {1 USD, 2000-05-03}
 Assets:Bar 2 STOCK.B {0.5 USD, 2000-05-03}\
'''

@pytest.mark.parametrize('text', [
    '2000-05-01 custom "autobean.stock_split" STOCK',
    '2000-05-01 custom "autobean.stock_split" 10',
    '2000-05-01 custom "autobean.stock_split"',
])
def test_invalid(text: str) -> None:
    _, errors = load(text)
    assert errors



import decimal
from typing import Any, Iterable, Iterator, Optional
from beancount.parser import options
from beancount.core import account_types
from beancount.core.data import Custom, Directive, Posting, Transaction
from beancount.core.amount import mul
from beancount.core import realization
from autobean.utils import plugin_lib


class Realizer:
    def __init__(self, options_map: dict[str, Any]) -> None:
        self._account_types = options.get_account_types(options_map)
        self._real_root = realization.RealAccount('')

    def realize_transaction(self, transaction: Transaction) -> None:
        for posting in transaction.postings:
            if not account_types.is_balance_sheet_account(posting.account, self._account_types):
                continue
            real_account = realization.get_or_create(self._real_root, posting.account)
            real_account.balance.add_position(posting)

    def get_split_postings(self, commodity: str, multiplier: decimal.Decimal) -> Iterator[Posting]:
        for real_account in realization.iter_children(self._real_root):
            for position in real_account.balance:
                if position.units.currency != commodity or not position.units.number:
                    continue
                new_cost = position.cost._replace(number=position.cost.number / multiplier)
                yield Posting(
                    account=real_account.account,
                    units=-position.units,
                    cost=position.cost,
                    price=None,
                    flag=None,
                    meta=None)
                yield Posting(
                    account=real_account.account,
                    units=mul(position.units, multiplier),
                    cost=new_cost,
                    price=None,
                    flag=None,
                    meta=None)


@plugin_lib.plugin('autobean.stock_split')
class Plugin(plugin_lib.BasePlugin):

    def process(self, entries: list[Directive], options: dict[str, Any], arg: Optional[str]) -> Iterable[Directive]:
        self._realizer = Realizer(options)
        return super().process(entries, options, arg)

    @plugin_lib.handle_custom('autobean.stock_split', 'exactly one multiplier and one commodity')
    def handle_stock_split(
            self,
            entry: Custom,
            multiplier: decimal.Decimal,
            commodity: plugin_lib.Currency,
    ) -> Iterator[Transaction]:
        postings = self._realizer.get_split_postings(commodity, multiplier)
        txn = Transaction(
            date=entry.date,
            flag='*',
            payee=None,
            narration=f'{commodity} split {multiplier}:1',
            tags=set(),
            links=set(),
            postings=list(postings),
            meta=entry.meta,
        )
        self._realizer.realize_transaction(txn)
        yield txn

    @plugin_lib.handle(Transaction)
    def handle_txn(self, txn: Transaction) -> Iterator[Transaction]:
        self._realizer.realize_transaction(txn)
        yield txn



import os.path
from autobean.include import plugin
import autobean.utils.plugin_test_utils as utils


@utils.generate_tests(os.path.dirname(__file__), plugin.IncludePlugin.plugin)
def test() -> None:
    pass


def test_include_option() -> None:
    test_path = os.path.join(os.path.dirname(__file__), 'simple')
    source_path = os.path.join(test_path, 'source.bean')
    external_path = os.path.join(test_path, '_external.bean')
    ledger = utils.load_ledger(source_path)

    # assumptions
    assert not ledger.errors
    assert 'include' in ledger.options
    assert external_path not in ledger.options['include']

    _, errors = utils.apply_plugin(
        plugin.IncludePlugin.plugin, ledger.entries, ledger.options, None)

    assert not errors, errors
    assert external_path in ledger.options['include'], 'included files not added into options["include"]'



import collections
import contextlib
import dataclasses
import decimal
import inspect
import re
import shlex
from typing import Any, Callable, ClassVar, Generic, Iterable, Iterator, NewType, Optional, Type, TypeVar, Union, get_args, get_origin
from beancount.core.data import Custom, Directive
from beancount.core.amount import Amount
from beancount.core import account as beancount_account
from beancount.parser import grammar
from autobean.utils import error_lib

Account = NewType('Account', str)
Currency = NewType('Currency', str)
_HANDLER_ATTR = '_autobean_handler'
_T = TypeVar('_T', bound=Directive)
_R = TypeVar('_R', bound=Iterable[Directive])
_Plugin = TypeVar('_Plugin', bound='BasePlugin')
_RegularHandlerImpl = Callable[[_Plugin, _T], _R]
_CustomHandlerImpl = Callable[..., _R]  # (self, custom, *args)


@dataclasses.dataclass(frozen=True)
class _RegularHandler(Generic[_Plugin, _T, _R]):
    directive_type: Type[Directive]
    when: Optional[Callable[[_Plugin], bool]]
    impl: _RegularHandlerImpl[_Plugin, _T, _R]


@dataclasses.dataclass(frozen=True)
class _CustomHandler:
    custom_type: str
    params_description: str
    params: list[inspect.Parameter]
    impl: _CustomHandlerImpl


class BasePlugin:
    _NAME: ClassVar[str]
    _ARGUMENT_TYPE: ClassVar[Any]
    _CUSTOM_SCOPE: ClassVar[Optional[re.Pattern]]
    _REGULAR_HANDLERS: ClassVar[dict[int, _RegularHandler]]
    _CUSTOM_HANDLERS: ClassVar[dict[str, list[_CustomHandler]]]

    _error_logger: error_lib.ErrorLogger

    def __init__(self) -> None:
        self._error_logger = error_lib.ErrorLogger()

    @classmethod
    def plugin(
            cls,
            entries: list[Directive],
            options: dict[str, Any],
            arg: Optional[str] = None,
    ) -> tuple[list[Directive], list[error_lib.Error]]:
        if cls._ARGUMENT_TYPE is None and arg is not None:
            raise ValueError(f'{cls._NAME} does not accept an argument')
        elif cls._ARGUMENT_TYPE is not None and arg is None:
            raise ValueError(f'{cls._NAME} expects an argument')
        inst = cls()
        return list(inst.process(entries, options, arg)), inst._error_logger.errors

    def process(
            self,
            entries: list[Directive],
            options: dict[str, Any],
            arg: Optional[str],
    ) -> Iterable[Directive]:
        if not self._REGULAR_HANDLERS and not self._CUSTOM_HANDLERS:
            return
        self._options = options
        self._argument = arg
        for entry in entries:
            yield from self._process_entry(entry)

    def _process_entry(self, entry: Directive) -> Iterator[Directive]:
        if isinstance(entry, Custom) and (handlers := self._CUSTOM_HANDLERS.get(entry.type)):
            for chandler in handlers:
                if (args := _get_args(entry.values or [], chandler)) is not None:
                    with _wrap_plugin_exception(entry, self._error_logger):
                        yield from chandler.impl(self, entry, *args)
                    return
            self._error_logger.log_error(
                error_lib.InvalidDirectiveError(
                    entry.meta,
                    f'Invalid arguments: {shlex.quote(entry.type)} expects {chandler.params_description}.',
                    entry))
            return
        if rhandler := self._REGULAR_HANDLERS.get(id(type(entry))):
            if rhandler.when is not None and not rhandler.when(self):
                return
            with _wrap_plugin_exception(entry, self._error_logger):
                yield from rhandler.impl(self, entry)
            return
        yield entry
        if isinstance(entry, Custom) and (
                self._CUSTOM_SCOPE is not None and self._CUSTOM_SCOPE.match(entry.type)):
            self._error_logger.log_error(
                error_lib.InvalidDirectiveError(
                    entry.meta, f'Unrecognized custom directive {shlex.quote(entry.type)}.', entry))
            return


@contextlib.contextmanager
def _wrap_plugin_exception(entry: Directive, error_logger: error_lib.ErrorLogger) -> Iterator[None]:
    try:
        yield
    except error_lib.PluginException as e:
        error_logger.log_error(error_lib.PluginError(
            e.meta or entry.meta, str(e), entry))


def _get_args(
        values: list[grammar.ValueType],
        handler: _CustomHandler,
) -> Optional[list[Any]]:
    queue = values[::-1]
    args = []
    for param in handler.params:
        if param.kind in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.POSITIONAL_ONLY):
            if (arg := _get_arg(param.annotation, queue)) is not None:
                args.append(arg)
            else:
                return None
        elif param.kind is inspect.Parameter.VAR_POSITIONAL:
            while queue:
                if (arg := _get_arg(param.annotation, queue)) is not None:
                    args.append(arg)
                else:
                    return None
    if queue:
        return None
    return args


def _get_arg(annotation: Any, queue: list[grammar.ValueType]) -> Optional[Any]:
    if not queue:
        return None
    value = queue.pop()
    if get_origin(annotation) is Union:
        options = get_args(annotation)
    else:
        options = (annotation,)
    for option in options:
        if value.dtype is option:
            return value.value
        elif option is Account:
            if value.dtype == beancount_account.TYPE:
                return Account(value.value)
        elif option is decimal.Decimal and value.dtype is Amount:
            if value.value.number is not None:
                queue.append(grammar.ValueType(Currency(value.value.currency), Currency))
                return value.value.number
    return None


def plugin(
        name: str,
        *,
        param_type: Any = None,
        custom_scope: Optional[str] = None,
) -> Callable[[Type[_Plugin]], Type[_Plugin]]:
    def decorator(cls: Type[_Plugin]) -> Type[_Plugin]:
        cls._NAME = name
        cls._ARGUMENT_TYPE = param_type
        cls._CUSTOM_SCOPE = re.compile(custom_scope) if custom_scope is not None else None
        regular_handlers = {}
        custom_handlers = collections.defaultdict(list) 
        for _, func in inspect.getmembers(cls, predicate=inspect.isfunction):
            handler = getattr(func, _HANDLER_ATTR, None)
            if isinstance(handler, _RegularHandler):
                regular_handlers[id(handler.directive_type)] = handler
            elif isinstance(handler, _CustomHandler):
                custom_handlers[handler.custom_type].append(handler)
        cls._REGULAR_HANDLERS = regular_handlers
        cls._CUSTOM_HANDLERS = custom_handlers
        return cls
    return decorator


def handle(
        directive_type: Type[_T],
        *,
        when: Optional[Callable[[_Plugin], bool]] = None,
) -> Callable[[_RegularHandlerImpl[_Plugin, _T, _R]], _RegularHandlerImpl[_Plugin, _T, _R]]:
    def decorator(impl: _RegularHandlerImpl[_Plugin, _T, _R]) -> _RegularHandlerImpl[_Plugin, _T, _R]:
        setattr(impl, _HANDLER_ATTR, _RegularHandler(directive_type, when, impl))
        return impl
    return decorator


def handle_custom(custom_type: str, params_description: str) -> Callable[[_CustomHandlerImpl], _CustomHandlerImpl]:
    def decorator(impl: _CustomHandlerImpl) -> _CustomHandlerImpl:
        setattr(impl, _HANDLER_ATTR, _CustomHandler(
            custom_type=custom_type,
            params_description=params_description,
            params=list(inspect.signature(impl).parameters.values())[2:],
            impl=impl))
        return impl
    return decorator



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

with open('requirements.txt', 'r') as f:
    requirements = list(filter(None, f.read().split('\n')))

setuptools.setup(
    name='autobean',
    version='0.2.0',
    author='SEIAROTg',
    author_email='seiarotg@gmail.com',
    description='A collection of plugins and scripts that help automating bookkeeping with beancount',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/SEIAROTg/autobean',
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Topic :: Office/Business :: Financial :: Accounting',
    ],
    python_requires='>=3.10',
)



from typing import Any, Iterable, NamedTuple, Optional
from beancount.loader import LoadError
from beancount.core.data import Directive, Meta


class Error(NamedTuple):
    source: Meta
    message: str
    entry: Directive | list[Directive] | None


class InvalidDirectiveError(Error):
    pass


class PluginError(Error):
    pass


class PluginException(Exception):
    def __init__(self, *args: Any, meta: Optional[Meta] = None, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.meta = meta


class ErrorLogger:
    errors: list[Error]

    def __init__(self) -> None:
        self.errors = []

    def log_error(self, error: Error) -> None:
        self.errors.append(error)

    def log_errors(self, errors: Iterable[Error]) -> None:
        self.errors.extend(errors)

    def log_loading_errors(self, errors: Iterable[Error], entry: Directive) -> None:
        for error in errors:
            if isinstance(error, LoadError):
                error = error._replace(
                    source=entry.meta,
                    entry=entry,
                )
            self.errors.append(error)



import os.path
import sys
from autobean.share import plugin
from autobean.utils import plugin_test_utils


@plugin_test_utils.generate_tests(os.path.dirname(__file__), plugin.Plugin.plugin)
def test() -> None:
    pass


if __name__ == '__main__':
    plugin_test_utils.generate_goldens(
        os.path.dirname(__file__), sys.argv[1], plugin.Plugin.plugin)



import collections
import dataclasses
import decimal
import functools
import itertools
from typing import Any, Iterator, Optional, TypeVar
from beancount.core import account as account_lib, amount as amount_lib, inventory as inventory_lib, convert, interpolate, realization
from beancount.core.data import Balance, Close, Custom, Directive, Open, Posting, Transaction
from beancount.core.amount import Amount
from beancount.core.position import Cost, CostSpec
from beancount.ops import balance as balance_lib
from autobean.utils import error_lib
from . import policy_lib, viewpoint_lib

# TODO: consider determining the tolerance in a better way
_PROPORTIONATE_TOLERANCE = decimal.Decimal(1e-6)
_O = TypeVar('_O', bound=policy_lib.Ownership)
_PostingPolicy = tuple[Posting, policy_lib.Policy[_O]]


def _amount_distrib(amount: Amount, weight: decimal.Decimal, total_weight: decimal.Decimal) -> Amount:
    return amount_lib.div(amount_lib.mul(amount, weight), total_weight)


def _costspec_distrib(costspec: CostSpec, weight: decimal.Decimal, total_weight: decimal.Decimal) -> CostSpec:
    number_total = costspec.number_total
    if number_total is not None:
        amount = Amount(number_total, costspec.currency)
        number_total = _amount_distrib(amount, weight, total_weight)
    return costspec._replace(
        number_total=number_total,
    )


def _posting_distrib(posting: Posting, weight: decimal.Decimal, total_weight: decimal.Decimal) -> Posting:
    units = _amount_distrib(posting.units, weight, total_weight)
    cost = posting.cost
    if isinstance(cost, CostSpec):
        cost = _costspec_distrib(cost, weight, total_weight)
    return posting._replace(
        units=units,
        cost=cost,
    )


def _split_posting_weighted(
        posting: Posting,
        ownership: policy_lib.WeightedOwnership,
) -> dict[str, Posting]:
    return {
        party: _posting_distrib(posting, weight, ownership.total_weight)
        for party, weight in ownership.weights.items()
    }


def _get_complement_posting(posting: Posting) -> Posting:
    units = convert.get_weight(posting)
    return posting._replace(units=units, price=None, cost=None)


@dataclasses.dataclass(frozen=True)
class _GroupedPostings:
    weighted: list[_PostingPolicy[policy_lib.WeightedOwnership]]
    prorated: list[_PostingPolicy[policy_lib.ProratedOwnership]]

    @classmethod
    def from_transaction(
            cls,
            transaction: Transaction,
            policy_db: policy_lib.PolicyDatabase,
    ) -> '_GroupedPostings':
        transaction_policy_def = policy_lib.try_parse_policy_definition(transaction.meta)
        policy_lib.strip_share_meta(transaction.meta)
        weighted_postings_policies = list[_PostingPolicy[policy_lib.WeightedOwnership]]()
        prorated_postings_policies = list[_PostingPolicy[policy_lib.ProratedOwnership]]()
        for posting in transaction.postings:
            policy = policy_db.get_posting_policy(posting, transaction_policy_def)
            if policy is None:
                raise error_lib.PluginException('No applicable share policy')
            if isinstance(policy.ownership, policy_lib.WeightedOwnership):
                weighted_postings_policies.append((posting, policy))
            elif isinstance(policy.ownership, policy_lib.ProratedOwnership):
                prorated_postings_policies.append((posting, policy))
            else:
                assert False
        return cls(weighted_postings_policies, prorated_postings_policies)


class _Inventory:

    def __init__(self) -> None:
        self._positions = collections.defaultdict[tuple[str, Amount, Cost], decimal.Decimal](
            decimal.Decimal)

    def add_posting(self, posting: Posting) -> None:
        key = (posting.units.currency, posting.price, posting.cost)
        self._positions[key] += posting.units.number

    def is_small(self, tolerances: decimal.Decimal | dict[str, decimal.Decimal]) -> bool:
        if isinstance(tolerances, decimal.Decimal):
            for balance in self._positions.values():
                if abs(balance) > tolerances:
                    return False
            return True
        for (currency, _, _), balance in self._positions.items():
            tolerance = tolerances.get(currency, decimal.Decimal(0))
            if abs(balance) > tolerance:
                return False
        return True

    @property
    def positions(self) -> dict[tuple[str, Amount, Cost], decimal.Decimal]:
        return self._positions


class _ConversionTableEntry:
    pass


@dataclasses.dataclass(frozen=True)
class _UnambiguousConversionTableEntry(_ConversionTableEntry):
    from_currency: str
    price: Optional[Amount]
    cost: Optional[Cost]

    @functools.cached_property
    def rate(self) -> decimal.Decimal:
        if self.cost:
            return self.cost.number
        elif self.price:
            return self.price.number
        assert False


class _AmbiguousConversionTableEntry(_ConversionTableEntry):
    pass


_AMBIGUOUS_CONVERSION_TABLE_ENTRY = _AmbiguousConversionTableEntry()


class _ConversionTable:
    def __init__(self, table: dict[str, _ConversionTableEntry]):
        self._table = table

    @classmethod
    def from_grouped_postings(cls, grouped_postings: _GroupedPostings) -> '_ConversionTable':
        entries_by_currency = collections.defaultdict[str, set[_UnambiguousConversionTableEntry]](set)
        no_conversion_currencies: set[str] = set()
        weighted: list[_PostingPolicy] = grouped_postings.weighted
        prorated: list[_PostingPolicy] = grouped_postings.prorated

        for posting, policy in itertools.chain(weighted, prorated):
            if not posting.price and not posting.cost:
                continue
            currency = posting.cost.currency if posting.cost else posting.price.currency
            if not policy.conversion:
                no_conversion_currencies.add(currency)
            entry = _UnambiguousConversionTableEntry(posting.units.currency, posting.price, posting.cost)
            entries_by_currency[currency].add(entry)

        table = dict[str, _ConversionTableEntry]()
        for currency in no_conversion_currencies:
            entries = entries_by_currency[currency]
            if len(entries) == 1:
                table[currency] = next(iter(entries))
            else:
                table[currency] = _AMBIGUOUS_CONVERSION_TABLE_ENTRY

        return cls(table)

    def create_complement_posting(
            self,
            account: str,
            number: decimal.Decimal,
            currency: str,
            price: Optional[Amount],
            cost: Optional[Cost],
            meta: Optional[dict[str, Any]],
    ) -> Posting:
        if entry := self._table.get(currency):
            if not isinstance(entry, _UnambiguousConversionTableEntry):
                raise error_lib.PluginException(
                    f'Ambiguous conversion for currency {currency}. '
                    'Consider not using share_conversion: FALSE.')
            units = Amount(
                number=number / entry.rate,
                currency=entry.from_currency)
            cost = entry.cost
            price = entry.price
        else:
            units = Amount(number, currency)
        return Posting(
            account=account,
            units=units,
            price=price,
            cost=cost,
            flag=None,
            meta=meta,
        )


class _ProratedOwnershipBuilder:
    def __init__(self) -> None:
        self._currency: Optional[str] = None
        self._weights = collections.defaultdict[str, decimal.Decimal](decimal.Decimal)
        self._total_weights = decimal.Decimal(0)

    def check_currency(self, currency: str) -> None:
        if self._currency is None:
            self._currency = currency
        elif self._currency != currency:
            raise error_lib.PluginException(
                f'Currency mismatch in prorated weights calculation: '
                f'{currency} != {self._currency}')

    def add_postings(self, postings: dict[str, Posting]) -> None:
        for party, posting in postings.items():
            self._weights[party] += posting.units.number
            self._total_weights += posting.units.number

    def build(self) -> policy_lib.WeightedOwnership:
        if not self._weights:
            raise error_lib.PluginException(
                f'Cannot determine prorated ownership weights with no participating postings')
        if not self._total_weights:
            raise error_lib.PluginException(
                f'Cannot determine prorated ownership weights with zero total weights')
        return policy_lib.WeightedOwnership(self._weights)


class _TransactionProcessor:
    def __init__(
            self,
            *,
            transaction: Transaction,
            policy_db: policy_lib.PolicyDatabase,
            options: dict[str, Any],
            receivable_account: str,
    ) -> None:
        self._transaction = transaction
        self._receivable_account = receivable_account
        self._tolerance = interpolate.infer_tolerances(self._transaction.postings, options)
        self._postings_by_party = collections.defaultdict[str, list[Posting]](list)
        self._inventory_by_party = collections.defaultdict[str, _Inventory](_Inventory)
        # complement receivable postings generated from explicit postings on receivables
        self._complement_receivables = collections.defaultdict[str, list[Posting]](list)

        grouped_postings = _GroupedPostings.from_transaction(transaction, policy_db)
        self._conversion_table = _ConversionTable.from_grouped_postings(grouped_postings)
        self._process_transaction(grouped_postings)

    def _add_weighted_posting(
            self,
            posting: Posting,
            ownership: policy_lib.WeightedOwnership,
    ) -> dict[str, Posting]:
        policy_lib.strip_share_meta(posting.meta)
        complement = _get_complement_posting(posting)
        party_postings = _split_posting_weighted(posting, ownership)
        for party, posting in party_postings.items():
            self._postings_by_party[party].append(posting)
        complement_party_postings = _split_posting_weighted(complement, ownership)
        for party, posting in complement_party_postings.items():
            self._inventory_by_party[party].add_posting(posting)
        parent, _, receivable_party = posting.account.rpartition(':')
        if parent == self._receivable_account:
            for party, posting in party_postings.items():
                self._complement_receivables[receivable_party].append(posting._replace(
                    account=f'{self._receivable_account}:{party}',
                    units=-posting.units,
                ))
        return party_postings

    def _process_transaction(
            self,
            grouped_postings: _GroupedPostings,
    ) -> None:
        prorated_ownership_builder = _ProratedOwnershipBuilder()
        for posting, weighted_policy in grouped_postings.weighted:
            party_postings = self._add_weighted_posting(posting, weighted_policy.ownership)
            if grouped_postings.prorated and weighted_policy.prorated_included:
                prorated_ownership_builder.check_currency(posting.units.currency)
                prorated_ownership_builder.add_postings(party_postings)
        if grouped_postings.prorated:
            prorated_ownership = prorated_ownership_builder.build()
            for posting, _ in grouped_postings.prorated:
                self._add_weighted_posting(posting, prorated_ownership)

    def realize(self, root: realization.RealAccount, accounts: set[str]) -> None:
        for party, postings in self._postings_by_party.items():
            for posting in postings:
                if posting.account in accounts:
                    real_account = realization.get_or_create(root, f'{posting.account}:{party}')
                    real_account.balance.add_position(posting)

    def get_postings(
            self,
            *,
            viewpoint: str,
            used_subaccounts: dict[str, set[str]],
    ) -> list[Posting]:
        if viewpoint == viewpoint_lib.NOBODY:
            return [
                *self._transaction.postings,
                *itertools.chain.from_iterable(self._complement_receivables.values()),
                *self._get_complement_postings(),
            ]
        if viewpoint == viewpoint_lib.EVERYONE:
            return [
                *self._get_split_postings(used_subaccounts=used_subaccounts),
                *itertools.chain.from_iterable(self._complement_receivables.values()),
                *self._get_complement_postings(),
            ]
        ret = [
            *self._postings_by_party[viewpoint],
            *self._complement_receivables[viewpoint],
        ]
        if self._postings_by_party[viewpoint]:
            ret += self._get_complement_postings(excluded_party=viewpoint)
        return ret

    def _get_split_postings(
            self,
            *,
            used_subaccounts: dict[str, set[str]],
    ) -> Iterator[Posting]:
        for party, postings in self._postings_by_party.items():
            for posting in postings:
                account = posting.account
                parent, _, _ = account.rpartition(':')
                if parent != self._receivable_account:
                    account = f'{posting.account}:[{party}]'
                    used_subaccounts[posting.account].add(account)
                    yield posting._replace(account=account)
                else:
                    yield posting

    def _get_complement_postings(
            self,
            *,
            excluded_party: Optional[str] = None) -> list[Posting]:
        return [
            self._conversion_table.create_complement_posting(
                account=f'{self._receivable_account}:{party}',
                number=number,
                currency=currency,
                price=price,
                cost=cost,
                meta=self._transaction.meta)
            for party, inventory in self._inventory_by_party.items()
            if party != excluded_party and not inventory.is_small(self._tolerance)
            for (currency, price, cost), number in inventory.positions.items()
        ]


class AccountSplitter:
    def __init__(
            self,
            policy_db: policy_lib.PolicyDatabase,
            options: dict[str, Any],
            viewpoint: str,
            asserted_accounts: set[str],
    ) -> None:
        self._policy_db = policy_db
        self._options = options
        self._viewpoint = viewpoint
        self._asserted_accounts = asserted_accounts
        self._real_root = realization.RealAccount('')
        self._used_subaccounts = collections.defaultdict[str, set[str]](set)

    def process_transaction(self, transaction: Transaction, receivable_account: str) -> Optional[Transaction]:
        processor = _TransactionProcessor(
            transaction=transaction,
            policy_db=self._policy_db,
            options=self._options,
            receivable_account=receivable_account)
        asserted_accounts = {
            posting.account
            for posting in transaction.postings
            if any(account in self._asserted_accounts for account in account_lib.parents(posting.account))
        }
        processor.realize(self._real_root, asserted_accounts)
        postings = processor.get_postings(
            viewpoint=self._viewpoint,
            used_subaccounts=self._used_subaccounts)
        if transaction.postings and not postings:
            # irrelevant to our viewpoint
            return None
        policy_lib.strip_share_meta(transaction.meta)
        return transaction._replace(postings=postings)

    def process_balance(self, balance: Balance, error_logger: error_lib.ErrorLogger) -> list[Balance]:
        if self._viewpoint == viewpoint_lib.NOBODY:
            policy_lib.strip_share_meta(balance.meta)
            return [balance]
        tolerance = balance_lib.get_balance_tolerance(balance, self._options)
        real_account = realization.get(self._real_root, balance.account)
        total_balance, balance_by_party = _compute_balance(real_account)
        _check_balance(total_balance, balance, balance.amount, tolerance, error_logger)
        policy = self._policy_db.get_balance_policy(balance)
        policy_lib.strip_share_meta(balance.meta)
        if not policy:
            return []
        if self._viewpoint == viewpoint_lib.EVERYONE:
            return [
                balance._replace(
                    account=f'{balance.account}:[{party}]',
                    amount=_amount_distrib(balance.amount, weight, policy.ownership.total_weight),
                )
                for party, weight in policy.ownership.weights.items()
            ]
        for party, weight in policy.ownership.weights.items():
            if party == self._viewpoint:
                continue  # will be checked by the returned balance directive
            balance_amount = _amount_distrib(balance.amount, weight, policy.ownership.total_weight)
            _check_balance(balance_by_party.get(party), balance, balance_amount, tolerance, error_logger)
        if self._viewpoint not in policy.ownership.weights:
            return []
        return [
            balance._replace(amount=_amount_distrib(
                balance.amount,
                policy.ownership.weights[self._viewpoint],
                policy.ownership.total_weight)),
        ]

    def process_proportionate(self, entry: Custom, account: str) -> Optional[Custom]:
        policy = self._policy_db.get_proportionate_policy(entry, account)
        if not policy:
            raise error_lib.PluginException(
                f'No applicable share policy found for autobean.share.proportionate on {account}')
        if len(policy.ownership.weights) > 1:
            # single party owner is by construction proportionate
            _check_proportionate(account, policy, self._real_root)
        if viewpoint_lib.is_overall(self._viewpoint):
            policy_lib.strip_share_meta(entry.meta)
            return entry
        return None

    def process_open_close(self, entries: list[Directive]) -> list[Directive]:
        if self._viewpoint != viewpoint_lib.EVERYONE:
            return entries
        results = []
        for entry in entries:
            if isinstance(entry, Open | Close):
                for subaccount in sorted(self._used_subaccounts.get(entry.account, (entry.account,))):
                    results.append(entry._replace(account=subaccount))
            else:
                results.append(entry)
        return results


def _compute_balance(real_account: Optional[realization.RealAccount]) -> tuple[
        inventory_lib.Inventory,
        dict[str, inventory_lib.Inventory]
]:
    total_balance = inventory_lib.Inventory()
    balance_by_party = collections.defaultdict[str, inventory_lib.Inventory](
        inventory_lib.Inventory)
    if real_account is not None:
        for ra in realization.iter_children(real_account, leaf_only=True):
            total_balance += ra.balance
            party = ra.account.split(':')[-1]
            balance_by_party[party] += ra.balance
    return total_balance, balance_by_party


def _check_balance(
        actual_balance: Optional[inventory_lib.Inventory],
        balance: Balance,
        expected_amount: Amount,
        tolerance: decimal.Decimal,
        error_logger: error_lib.ErrorLogger,
) -> None:
    if actual_balance is None:
        actual_amount = amount_lib.Amount(decimal.Decimal(0), expected_amount.currency)
    else:
        actual_amount = actual_balance.get_currency_units(expected_amount.currency)
    diff_amount = amount_lib.sub(actual_amount, expected_amount)
    if abs(diff_amount.number) > tolerance:
        diff_direction = 'too much' if diff_amount.number > 0 else 'too little'
        error_logger.log_error(balance_lib.BalanceError(
            balance.meta,
            f'Balance failed for {balance.account!r}: '
            f'expected {expected_amount} != accumulated {actual_amount} '
            f'({abs(diff_amount.number)} {diff_direction})',
            balance,
        ))


def _check_proportionate(
        account: str,
        policy: policy_lib.Policy[policy_lib.WeightedOwnership],
        real_root: realization.RealAccount,
) -> None:
    real_account = realization.get(real_root, account)
    if real_account is None:
        return  # empty account is by definition proportionate
    total_balance, balance_by_party = _compute_balance(real_account)
    if set(balance_by_party) - set(policy.ownership.weights):
        raise error_lib.PluginException(f'Disproportionate balance on account {account}')
    for key, position in total_balance.items():
        for party, weight in policy.ownership.weights.items():
            expected_num = position.units.number * weight / policy.ownership.total_weight
            actual_num = 0
            if (party_balance := balance_by_party.get(party)) is not None:
                if party_position := party_balance.get(key):
                    actual_num = party_position.units.number
            diff_num = actual_num - expected_num
            if abs(diff_num) > _PROPORTIONATE_TOLERANCE:
                raise error_lib.PluginException(f'Disproportionate balance on account {account}')
